import torch
from torch.nn import Dropout, Linear
from torch.nn.functional import cross_entropy
from torch.nn.utils import clip_grad_norm_
from transformers import AutoModel, AutoTokenizer


def data_cuda(*tensors, device='gpu', non_blocking=False):
    if device == 'gpu':
        cuda_tensors = []
        for tensor in tensors:
            cuda_tensors.append(tensor.cuda(non_blocking=non_blocking))
    else:
        cuda_tensors = tensors

    if len(cuda_tensors) > 1:
        return cuda_tensors
    else:
        return cuda_tensors[0]


class BERTNLI(torch.nn.Module):
    LABEL_CONTRADICTION = 2
    LABEL_ENTAILMENT = 0
    LABEL_NEUTRAL = 1

    def __init__(self, name_or_path, bert_type='bert', cls='linear', length=128, force_lowercase=False, device='cpu',
                 verbose=False):
        super(BERTNLI, self).__init__()
        self.bert_type = bert_type
        self.cls = cls
        self.force_lowercase = force_lowercase
        self.bert = AutoModel.from_pretrained(name_or_path)
        self.tokenizer = AutoTokenizer.from_pretrained(name_or_path)
        self.dropout = Dropout(0.1)
        self.linear = Linear(self.bert.config.hidden_size, 3)
        self.length = length
        self.device = device
        self.verbose = verbose

    @classmethod
    def train_step(cls, logits, gold_labels, optimizer, model=None, grad_clip=None):
        optimizer.zero_grad()
        target = logits.new_zeros((logits.shape[0],), dtype=torch.long)
        for i, gold_label in enumerate(gold_labels):
            if gold_label == 'entailment':
                target[i] = cls.LABEL_ENTAILMENT
            elif gold_label == 'neutral':
                target[i] = cls.LABEL_NEUTRAL
            elif gold_label == 'contradiction':
                target[i] = cls.LABEL_CONTRADICTION
            else:
                raise ValueError('Unknown label {0}'.format(gold_label))
        loss = cross_entropy(logits, target)
        loss.backward()
        if grad_clip is not None:
            clip_grad_norm_(model.parameters(), grad_clip)
        optimizer.step()
        return float(loss.detach().cpu())

    def cuda(self, device=None):
        super(BERTNLI, self).cuda(device)
        if device != torch.device('cpu'):
            self.device = 'gpu'

    def forward(self, sent1s, sent2s):
        buffer, boundaries, max_len = [], [], 0
        for sent1, sent2 in zip(sent1s, sent2s):
            if self.force_lowercase:
                sent1 = sent1.lower()
                sent2 = sent2.lower()
            toks1 = self.tokenizer.tokenize(sent1)
            toks2 = self.tokenizer.tokenize(sent2)
            tokens = ['[CLS]'] + toks1 + ['[SEP]'] + toks2 + ['[SEP]']
            buffer.append(tokens)
            boundaries.append(len(toks1) + 2)
            if len(tokens) > max_len:
                max_len = len(tokens)
        if max_len > self.length:
            max_len = self.length
        token_ids, attn_mask = [], []
        seg_ids = [] if self.bert_type != 'distilbert' else None
        for idx, tokens in enumerate(buffer):
            if len(tokens) < max_len:
                for _ in range(max_len - len(tokens)):
                    tokens.append('[PAD]')
            elif len(tokens) > max_len:
                if self.verbose:
                    print('Truncating pair from {0}->{1}'.format(len(tokens), max_len))
                tokens = tokens[:max_len]
            attn_mask.append(torch.tensor([1 if token != '[PAD]' else 0 for token in tokens]))
            token_ids.append(torch.tensor(self.tokenizer.convert_tokens_to_ids(tokens)))
            if seg_ids is not None:
                seg_ids.append(torch.tensor([0 if i < boundaries[idx] else 1 for i in range(len(tokens))]))
        token_ids = torch.stack(token_ids, dim=0)
        attn_mask = torch.stack(attn_mask, dim=0)
        if seg_ids is not None:
            seg_ids = torch.stack(seg_ids, dim=0)
            token_ids, attn_mask, seg_ids = data_cuda(token_ids, attn_mask, seg_ids, device=self.device)
        else:
            token_ids, attn_mask = data_cuda(token_ids, attn_mask, device=self.device)
        out = self.bert(token_ids.cuda(), attention_mask=attn_mask.cuda(), token_type_ids=seg_ids.cuda())
        reps = out["last_hidden_state"]
        cls = out["pooler_output"]
        if self.cls == 'token':
            reps = reps[:, 0]
            reps = self.dropout(reps)
            return self.linear(reps)
        else:
            cls = self.dropout(cls)
            return self.linear(cls)



import torch.nn as nn
from vilmedic.models.utils import get_n_params
import functools
import os
from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.decoder.evaluation import evaluation as evaluation_
import glob

from .RRG import RRG
import numpy as np
import copy
import torch
from vilmedic.blocks.rl.SCST import SCST


def evaluation(models, config, dl, **kwargs):
    models = [m.model if not isinstance(m, nn.DataParallel) else m.module.model for m in
              models]  # Get trained RRG instance
    return evaluation_(models, config, dl, **kwargs)


def get_ckpt(ckpt):
    if ".pth" in ckpt:
        return ckpt
    elif os.path.isdir(ckpt):
        ckpt = glob.glob(os.path.join(ckpt, "*.pth"))
        assert len(ckpt) == 1
        return ckpt
    else:
        raise FileNotFoundError(ckpt)


class RRG_SCST(nn.Module):

    def __init__(self, decoder, cnn, dl, scores="ROUGEL", ckpt=None, scores_args=None, scores_weights=None, top_k=None,
                 use_nll=False, **kwargs):
        super().__init__()

        # Models

        self.model = RRG(copy.deepcopy(decoder), copy.deepcopy(cnn), dl=dl, **kwargs)
        if ckpt:
            state_dict = torch.load(get_ckpt(ckpt))["model"]
            state_dict = {k.replace('module.', ''): v for k, v in state_dict.items()}
            self.model.load_state_dict(state_dict, strict=True)

        # SCST
        self.scst = SCST(decoder=self.model.dec.decoder,
                         dl=dl,
                         scores=scores,
                         scores_args=scores_args,
                         scores_weights=scores_weights,
                         use_nll=use_nll,
                         top_k=top_k)

        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, images_mask=None, encoder_outputs=None, **kwargs):
        # 1 Greedy
        with torch.no_grad():
            self.model.eval()
            encoder_hidden_states, encoder_attention_mask = self.model.encode(images.cuda(), images_mask, **kwargs)
            reward_greedy, greedy_hyp_list, ref_list = self.scst.forward_greedy(
                input_ids=input_ids,
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask
            )

        # 2. Sampling
        self.model.train()
        encoder_hidden_states, encoder_attention_mask = self.model.encode(images.cuda(), images_mask, **kwargs)
        loss, delta_reward, delta_reward_per_metric, reward_sampling, sampling_hyp_list = self.scst.forward_sampling(
            input_ids=input_ids,
            attention_mask=attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            reward_greedy=reward_greedy
        )

        return {"loss": loss,
                "custom_print": "reward_sampling {}, "
                                "delta_reward: {}".format(torch.mean(torch.tensor(reward_sampling)),
                                                          torch.mean(torch.tensor(delta_reward))),
                }

    def __repr__(self):
        s = "RRG_SCST\n"
        s += str(self.scst) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
from torch import nn, einsum
from einops import rearrange
from einops.layers.torch import Rearrange, Reduce
import torch.nn.functional as F

# helpers

def exists(val):
    return val is not None

def default(val, d):
    return val if exists(val) else d

def cast_tuple(val, length = 1):
    return val if isinstance(val, tuple) else ((val,) * length)

def divisible_by(val, d):
    return (val % d) == 0

# helper classes

class Downsample(nn.Module):
    def __init__(self, dim_in, dim_out):
        super().__init__()
        self.conv = nn.Conv2d(dim_in, dim_out, 3, stride = 2, padding = 1)

    def forward(self, x):
        return self.conv(x)

class PEG(nn.Module):
    def __init__(self, dim, kernel_size = 3):
        super().__init__()
        self.proj = nn.Conv2d(dim, dim, kernel_size = kernel_size, padding = kernel_size // 2, groups = dim, stride = 1)

    def forward(self, x):
        return self.proj(x) + x

# transformer classes

def FeedForward(dim, mult = 4, dropout = 0.):
    return nn.Sequential(
        nn.LayerNorm(dim),
        nn.Linear(dim, dim * mult, 1),
        nn.GELU(),
        nn.Dropout(dropout),
        nn.Linear(dim * mult, dim, 1)
    )

class Attention(nn.Module):
    def __init__(
        self,
        dim,
        heads = 4,
        dim_head = 32,
        dropout = 0.
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head ** -0.5
        inner_dim = dim_head * heads

        self.norm = nn.LayerNorm(dim)
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias = False)
        self.to_out = nn.Linear(inner_dim, dim)

    def forward(self, x, rel_pos_bias = None):
        h = self.heads

        # prenorm

        x = self.norm(x)

        q, k, v = self.to_qkv(x).chunk(3, dim = -1)

        # split heads

        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
        q = q * self.scale

        sim = einsum('b h i d, b h j d -> b h i j', q, k)

        # add relative positional bias for local tokens

        if exists(rel_pos_bias):
            sim = sim + rel_pos_bias

        attn = sim.softmax(dim = -1)

        # merge heads

        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        return self.to_out(out)

class R2LTransformer(nn.Module):
    def __init__(
        self,
        dim,
        *,
        window_size,
        depth = 4,
        heads = 4,
        dim_head = 32,
        attn_dropout = 0.,
        ff_dropout = 0.,
    ):
        super().__init__()
        self.layers = nn.ModuleList([])

        self.window_size = window_size
        rel_positions = 2 * window_size - 1
        self.local_rel_pos_bias = nn.Embedding(rel_positions ** 2, heads)

        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                Attention(dim, heads = heads, dim_head = dim_head, dropout = attn_dropout),
                FeedForward(dim, dropout = ff_dropout)
            ]))

    def forward(self, local_tokens, region_tokens):
        device = local_tokens.device
        lh, lw = local_tokens.shape[-2:]
        rh, rw = region_tokens.shape[-2:]
        window_size_h, window_size_w = lh // rh, lw // rw

        local_tokens = rearrange(local_tokens, 'b c h w -> b (h w) c')
        region_tokens = rearrange(region_tokens, 'b c h w -> b (h w) c')

        # calculate local relative positional bias
        
        h_range = torch.arange(window_size_h, device = device)
        w_range = torch.arange(window_size_w, device = device)

        grid_x, grid_y = torch.meshgrid(h_range, w_range)
        grid = torch.stack((grid_x, grid_y))
        grid = rearrange(grid, 'c h w -> c (h w)')
        grid = (grid[:, :, None] - grid[:, None, :]) + (self.window_size - 1)
        bias_indices = (grid * torch.tensor([1, self.window_size * 2 - 1], device = device)[:, None, None]).sum(dim = 0)
        rel_pos_bias = self.local_rel_pos_bias(bias_indices)
        rel_pos_bias = rearrange(rel_pos_bias, 'i j h -> () h i j')
        rel_pos_bias = F.pad(rel_pos_bias, (1, 0, 1, 0), value = 0)

        # go through r2l transformer layers

        for attn, ff in self.layers:
            region_tokens = attn(region_tokens) + region_tokens

            # concat region tokens to local tokens

            local_tokens = rearrange(local_tokens, 'b (h w) d -> b h w d', h = lh)
            local_tokens = rearrange(local_tokens, 'b (h p1) (w p2) d -> (b h w) (p1 p2) d', p1 = window_size_h, p2 = window_size_w)
            region_tokens = rearrange(region_tokens, 'b n d -> (b n) () d')

            # do self attention on local tokens, along with its regional token

            region_and_local_tokens = torch.cat((region_tokens, local_tokens), dim = 1)
            region_and_local_tokens = attn(region_and_local_tokens, rel_pos_bias = rel_pos_bias) + region_and_local_tokens

            # feedforward

            region_and_local_tokens = ff(region_and_local_tokens) + region_and_local_tokens

            # split back local and regional tokens

            region_tokens, local_tokens = region_and_local_tokens[:, :1], region_and_local_tokens[:, 1:]
            local_tokens = rearrange(local_tokens, '(b h w) (p1 p2) d -> b (h p1 w p2) d', h = lh // window_size_h, w = lw // window_size_w, p1 = window_size_h)
            region_tokens = rearrange(region_tokens, '(b n) () d -> b n d', n = rh * rw)

        local_tokens = rearrange(local_tokens, 'b (h w) c -> b c h w', h = lh, w = lw)
        region_tokens = rearrange(region_tokens, 'b (h w) c -> b c h w', h = rh, w = rw)
        return local_tokens, region_tokens

# classes

class RegionViT(nn.Module):
    def __init__(
        self,
        *,
        dim = (64, 128, 256, 512),
        depth = (2, 2, 8, 2),
        window_size = 7,
        num_classes = 1000,
        tokenize_local_3_conv = False,
        local_patch_size = 4,
        use_peg = False,
        attn_dropout = 0.,
        ff_dropout = 0.,
        channels = 3,
    ):
        super().__init__()
        dim = cast_tuple(dim, 4)
        depth = cast_tuple(depth, 4)
        assert len(dim) == 4, 'dim needs to be a single value or a tuple of length 4'
        assert len(depth) == 4, 'depth needs to be a single value or a tuple of length 4'

        self.local_patch_size = local_patch_size

        region_patch_size = local_patch_size * window_size
        self.region_patch_size = local_patch_size * window_size

        init_dim, *_, last_dim = dim

        # local and region encoders

        if tokenize_local_3_conv:
            self.local_encoder = nn.Sequential(
                nn.Conv2d(3, init_dim, 3, 2, 1),
                nn.LayerNorm(init_dim),
                nn.GELU(),
                nn.Conv2d(init_dim, init_dim, 3, 2, 1),
                nn.LayerNorm(init_dim),
                nn.GELU(),
                nn.Conv2d(init_dim, init_dim, 3, 1, 1)
            )
        else:
            self.local_encoder = nn.Conv2d(3, init_dim, 8, 4, 3)

        self.region_encoder = nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (c p1 p2) h w', p1 = region_patch_size, p2 = region_patch_size),
            nn.Conv2d((region_patch_size ** 2) * channels, init_dim, 1)
        )

        # layers

        current_dim = init_dim
        self.layers = nn.ModuleList([])

        for ind, dim, num_layers in zip(range(4), dim, depth):
            not_first = ind != 0
            need_downsample = not_first
            need_peg = not_first and use_peg

            self.layers.append(nn.ModuleList([
                Downsample(current_dim, dim) if need_downsample else nn.Identity(),
                PEG(dim) if need_peg else nn.Identity(),
                R2LTransformer(dim, depth = num_layers, window_size = window_size, attn_dropout = attn_dropout, ff_dropout = ff_dropout)
            ]))

            current_dim = dim

        # final logits

        self.to_logits = nn.Sequential(
            Reduce('b c h w -> b c', 'mean'),
            nn.LayerNorm(last_dim),
            nn.Linear(last_dim, num_classes)
        )

    def forward(self, x):
        *_, h, w = x.shape
        assert divisible_by(h, self.region_patch_size) and divisible_by(w, self.region_patch_size), 'height and width must be divisible by region patch size'
        assert divisible_by(h, self.local_patch_size) and divisible_by(w, self.local_patch_size), 'height and width must be divisible by local patch size'

        local_tokens = self.local_encoder(x)
        region_tokens = self.region_encoder(x)

        for down, peg, transformer in self.layers:
            local_tokens, region_tokens = down(local_tokens), down(region_tokens)
            local_tokens = peg(local_tokens)
            local_tokens, region_tokens = transformer(local_tokens, region_tokens)

        return self.to_logits(region_tokens)



import torch
import torch.nn as nn
from vilmedic.models.utils import get_n_params

from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.decoder.decoder_model import DecoderModel
from vilmedic.blocks.huggingface.decoder.evaluation import evaluation
from einops import rearrange


class RRG(nn.Module):

    def __init__(self, decoder, cnn, dl=None, **kwargs):
        super().__init__()
        # Decoder
        if dl:
            decoder.vocab_size = dl.dataset.seq.tokenizer.vocab_size
        self.dec = DecoderModel(decoder)

        # Encoder
        visual_embedding_dim = cnn.pop("visual_embedding_dim", None)
        cnn = eval(cnn.pop('proto'))(**cnn)
        if visual_embedding_dim:
            visual_projection = nn.Linear(visual_embedding_dim, self.dec.decoder.config.hidden_size)
        else:
            visual_projection = nn.Identity()

        self.enc = nn.Sequential(cnn, visual_projection)

        # Evaluation
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, images_mask=None, encoder_outputs=None,
                encoder_attention_mask=None, epoch=None, iteration=None, **kwargs):

        if torch.cuda.is_available():
            input_ids = input_ids.cuda()
            attention_mask = attention_mask.cuda()

        if encoder_outputs is None:
            encoder_outputs, encoder_attention_mask = self.encode(images, images_mask, **kwargs)

        out = self.dec(input_ids=input_ids,
                       attention_mask=attention_mask,
                       encoder_outputs=encoder_outputs,
                       encoder_attention_mask=encoder_attention_mask,
                       **kwargs)

        return out

    # Necessary for generation
    def encode(self, images, images_mask=None, **kwargs):
        if torch.cuda.is_available():
            images = images.cuda()

        # Single-image forward pass
        if len(images.shape) == 4:
            feature = self.enc(images)
            feature_mask = (torch.sum(torch.abs(feature), dim=-1) != 0)
            return feature, feature_mask

        assert len(images.shape) == 5, "wrong images shape"
        # Multi-image or multi patch:

        # 3D encoder
        if self.enc[0].is3D:
            features = self.enc(images)
            feature_mask = (torch.sum(torch.abs(features), dim=-1) != 0)
            return features, feature_mask

        # Multi forward pass
        images = rearrange(images, 'd0 d1 d2 d3 d4 -> (d0 d1) d2 d3 d4')
        feature = self.enc(images)

        # Masking features of empty images
        num_images = images.shape[1]
        feature = feature.view(int(feature.shape[0] / num_images), num_images, feature.shape[-2], feature.shape[-1])

        if torch.cuda.is_available() and images_mask is not None:
            images_mask = images_mask.cuda()

        if images_mask is not None:
            feature = feature * images_mask.unsqueeze(-1).unsqueeze(-1)

        if torch.cuda.is_available():
            feature = feature.cuda()

        # Creating feature-wise attention mask
        feature = rearrange(feature, 'd0 d1 d2 d3 -> d0 (d1 d2) d3')
        feature_mask = (torch.sum(torch.abs(feature), dim=-1) != 0)
        return feature, feature_mask

    def __repr__(self):
        s = "model: RRG\n"
        s += "(enc):" + str(self.enc) + '\n'
        s += "(dec):" + str(self.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import tqdm
import torch
import functools
from .beam_search import beam_search
import torch.nn as nn


def get_special_token_ids(model, tokenizer):
    bos_token_id = model.config.bos_token_id
    eos_token_id = model.config.eos_token_id
    pad_token_id = model.config.pad_token_id
    if None in [bos_token_id, eos_token_id, pad_token_id]:
        bos_token_id = tokenizer.vocab[tokenizer.cls_token]
        eos_token_id = tokenizer.vocab[tokenizer.sep_token]
        pad_token_id = tokenizer.vocab[tokenizer.pad_token]

    return bos_token_id, eos_token_id, pad_token_id


def evaluation(models, config, dl, **kwargs):
    models = [m if not isinstance(m, nn.DataParallel) else m.module for m in models]
    hf_models = [model.dec.decoder for model in models]

    # We are in an ensembling scenario, we override huggingface beam-search function
    hf_models[0].beam_search = functools.partial(beam_search, hf_models[0])

    # Get tokenizer and reference sentences from dataloader
    try:
        ref_str = 'input_ids'
        tokenizer = dl.dataset.tokenizer
        max_len = dl.dataset.tokenizer_max_len
    except AttributeError:
        ref_str = 'decoder_input_ids'
        tokenizer = dl.dataset.tgt_tokenizer
        max_len = dl.dataset.tgt_tokenizer_max_len

    # Get tokens
    bos_token_id, eos_token_id, pad_token_id = get_special_token_ids(hf_models[0], tokenizer)

    ref_list = []
    hyp_list = []

    with torch.no_grad():
        for batch in tqdm.tqdm(dl):
            batch = {k: v.cuda() if isinstance(v, torch.Tensor) else v for k, v in batch.items()}
            # Expanding inputs
            batch_size = batch[ref_str].shape[0]
            expanded_return_idx = (
                torch.arange(batch_size).view(-1, 1).repeat(1, config.beam_width).view(-1).cuda()
            )
            # Getting encoder infos
            encoder_outputs = []
            encoder_attention_masks = []
            for hf in models:
                encoder_output, encoder_attention_mask = hf.encode(**batch)
                encoder_outputs.append(encoder_output)
                encoder_attention_masks.append(encoder_attention_mask)

            # Registering models and encoder outputs
            model_kwargs = {
                "encoders_outputs":
                    [
                        {"encoder_hidden_states": output.index_select(0, expanded_return_idx),
                         "encoder_attention_mask": mask.index_select(0, expanded_return_idx)
                         } for output, mask in zip(encoder_outputs, encoder_attention_masks)
                    ],
                "hf_models": hf_models
            }

            # lets gooooo
            hyps = hf_models[0].generate(
                input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * bos_token_id,
                num_return_sequences=1,
                max_length=max_len,
                num_beams=config.beam_width,
                length_penalty=config.length_penalty,
                bos_token_id=bos_token_id,
                eos_token_id=eos_token_id,
                pad_token_id=pad_token_id,
                use_cache=True,
                **model_kwargs
            )

            refs = batch[ref_str]
            for h, r in zip(hyps, refs):
                hyp_list.append(tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
                ref_list.append(tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))
            # break
        return {'refs': ref_list, 'hyps': hyp_list}



import os
import torch
import time
import numpy as np
import torch.nn as nn
from torchmetrics.functional.text.bert import BERTScorer


class BertScore(nn.Module):
    def __init__(self):
        super(BertScore, self).__init__()
        with torch.no_grad():
            self.bert_scorer = BERTScorer(model_type='distilbert-base-uncased',
                                          num_layers=5,
                                          batch_size=64,
                                          nthreads=4,
                                          all_layers=False,
                                          idf=False,
                                          device='cuda',
                                          lang='en',
                                          rescale_with_baseline=True,
                                          baseline_path=None)

    def forward(self, refs, hyps):
        p, r, f = self.bert_scorer.score(
            cands=hyps,
            refs=refs,
            verbose=False,
            batch_size=64,
        )
        return torch.mean(f).item(), f.tolist()


if __name__ == '__main__':
    x, y = (BertScore()(
        hyps=[
            "nothing to do lol",
            "nothing to do x",
            'there are moderate bilateral pleural effusions with overlying atelectasis,  underlying consolidation not excluded. mild prominence of the interstitial  markings suggests mild pulmonary edema. the cardiac silhouette is mildly  enlarged. the mediastinal contours are unremarkable. there is no evidence of  pneumothorax.'
        ],
        refs=[
            'heart size is moderately enlarged. the mediastinal and hilar contours are unchanged. there is no pulmonary edema. small left pleural effusion is present. patchy opacities in the lung bases likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.',
            'heart size is mildly enlarged. the mediastinal and hilar contours are normal. there is mild pulmonary edema. moderate bilateral pleural effusions are present, left greater than right. bibasilar airspace opacities likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.',
            'heart size is mildly enlarged. the mediastinal and hilar contours are normal. there is mild pulmonary edema. moderate bilateral pleural effusions are present, left greater than right. bibasilar airspace opacities likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.'
        ])
    )
    print(x)
    print(y)



import gdown
import os
import zipfile
from tqdm import tqdm
from huggingface_hub import hf_hub_download, list_repo_files


def edit_vocab_path_in_dict(obj, keys, replace_value):
    for k, v in obj.items():
        if isinstance(v, dict):
            obj[k] = edit_vocab_path_in_dict(v, keys, replace_value)
    for key in keys:
        if key in obj:
            obj[key] = os.path.join(replace_value, obj[key])
    return obj


def download_images(data_name, file_id, unzip_dir):
    zip_name = os.path.join("data", data_name) + ".zip"
    target_dir = os.path.join(unzip_dir, data_name)

    if not os.path.exists(target_dir):
        gdown.download(url="https://drive.google.com/uc?id=" + file_id,
                       output=zip_name,
                       quiet=False)
        print("Unzipping...")
        with zipfile.ZipFile(zip_name, 'r') as zf:
            for member in tqdm(zf.infolist(), desc='Extracting '):
                zf.extract(member, unzip_dir)

        os.remove(zip_name)
        return
    print('{} already exists'.format(target_dir))


def download_model(repo_id, cache_dir, filename=None):
    # creating cache_dir
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir, exist_ok=True)

    # is HuggingFace repo ?
    if '/' in repo_id:
        # Single file or whole repo?
        if filename is not None:
            files = [filename]
        else:
            files = list(set(list_repo_files(repo_id=repo_id)).difference({'README.md', '.gitattributes'}))

        # Download
        for f in files:
            try:
                hf_hub_download(repo_id=repo_id, filename=f, cache_dir=cache_dir, force_filename=f)
            except Exception as e:
                print(e)

    else:  # Otherwise gdrive, full repo
        gdown.download_folder(id=repo_id,
                              output=cache_dir,
                              quiet=False)


def download_data(file_id, unzip_dir):
    if not os.path.exists(unzip_dir):
        os.makedirs(unzip_dir, exist_ok=True)
    gdown.download_folder(id=file_id,
                          output=unzip_dir,
                          quiet=False)



import torch.nn as nn
import torch
from collections import OrderedDict
from vilmedic.models.utils import get_n_params

from vilmedic.blocks.huggingface.decoder.decoder_model import DecoderModel
from vilmedic.blocks.huggingface.encoder.encoder_model import EncoderModel
from vilmedic.blocks.huggingface.decoder.evaluation import evaluation


class RRS(nn.Module):

    def __init__(self, encoder, decoder, dl, **kwargs):
        super().__init__()

        # Encoder
        encoder.vocab_size = dl.dataset.src.tokenizer.vocab_size
        self.enc = EncoderModel(encoder)

        # Decoder
        decoder.vocab_size = dl.dataset.tgt.tokenizer.vocab_size
        self.dec = DecoderModel(decoder)

        # Evaluation
        self.eval_func = evaluation

        # Tokenizer
        self.tokenizer = dl.dataset.tgt_tokenizer

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask, encoder_outputs=None,
                encoder_attention_mask=None,
                epoch=None, iteration=None,
                **kwargs):

        if torch.cuda.is_available():
            input_ids = input_ids.cuda()
            attention_mask = attention_mask.cuda()

        if encoder_outputs is None:
            encoder_outputs, encoder_attention_mask = self.encode(input_ids, attention_mask, **kwargs)

        out = self.dec(input_ids=decoder_input_ids,
                       attention_mask=decoder_attention_mask,
                       encoder_outputs=encoder_outputs,
                       encoder_attention_mask=encoder_attention_mask,
                       **kwargs)

        return out

    def encode(self, input_ids, attention_mask, **kwargs):
        encoder_outputs = self.enc(input_ids, attention_mask, return_dict=True)
        return encoder_outputs.last_hidden_state, attention_mask

    def __repr__(self):
        s = "model: RRS\n"
        s += "(enc):" + str(self.enc) + '\n'
        s += "(dec):" + str(self.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import os
import torch
import time
import numpy as np
import torch.nn as nn
from vilmedic.blocks.scorers import RadEntityMatchExact
from vilmedic.blocks.scorers.RadEntityNLI.nli import SimpleNLI
from vilmedic.constants import EXTRA_CACHE_DIR
from vilmedic.zoo.utils import download_model

from torchmetrics.functional.text.bert import BERTScorer
from itertools import chain, product


class RadEntityNLI(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        # NER types
        self.target_types = {'S-ANATOMY', 'S-OBSERVATION'}
        self.match_exact = RadEntityMatchExact()

        # Downloading pretrain model from huggingface
        checkpoint = os.path.join(EXTRA_CACHE_DIR, "model_medrad_19k.gz")
        if not os.path.exists(checkpoint):
            download_model(repo_id='StanfordAIMI/RRG_scorers', cache_dir=EXTRA_CACHE_DIR,
                           filename="model_medrad_19k.gz")

        # NLI scorer
        model = SimpleNLI.load_model(checkpoint)
        self.nli = SimpleNLI(model, batch=24, neutral_score=0.3333333333333333, nthreads=2, pin_memory=False,
                             bert_score='distilbert-base-uncased', cache=200000,
                             verbose=False)

        # BertSore
        self.bert_scorer = BERTScorer(model_type='distilbert-base-uncased',
                                      num_layers=5,
                                      batch_size=64,
                                      nthreads=4,
                                      all_layers=False,
                                      idf=False,
                                      device='cuda',
                                      lang='en',
                                      rescale_with_baseline=True,
                                      baseline_path=None)

    def forward(self, refs, hyps):
        t = time.time()
        with torch.no_grad():
            # # Getting radiology reports with rad entities
            _, _, docs_h, docs_r = self.match_exact(refs, hyps)

            scores_e = []
            for doc_h, doc_r in zip(docs_h, docs_r):

                hyp_report = [' '.join([token['text'] for token in sentence.to_dict()]) for sentence in
                              doc_h.sentences]
                ref_report = [' '.join([token['text'] for token in sentence.to_dict()]) for sentence in
                              doc_r.sentences]

                if len(hyp_report) == 0 or len(ref_report) == 0:
                    continue

                ner_h = [[ner['text'] for ner in sentence.to_dict() if ner['ner'] in self.target_types] for sentence in
                         doc_h.sentences]
                ner_r = [[ner['text'] for ner in sentence.to_dict() if ner['ner'] in self.target_types] for sentence in
                         doc_r.sentences]

                # getting all sentence pairs scores
                pairs = list(product(hyp_report, ref_report))
                _, _, f_scores = self.bert_scorer.score(
                    cands=[p[0] for p in pairs],
                    refs=[p[1] for p in pairs],
                    verbose=False,
                    batch_size=64,
                )

                f_scores = torch.reshape(torch.tensor(f_scores), (len(hyp_report), len(ref_report)))

                # self.bert_scorer.plot_example([p[0] for p in pairs][0], [p[1] for p in pairs][0])

                # precision
                match_p = 0
                entity_ner_r = list(chain.from_iterable(ner_r))
                total_p = 0
                for hyp_sentence, hyp_sentence_entities, hyp_f_score in zip(hyp_report, ner_h, f_scores):
                    # No entites in current sentence
                    if not hyp_sentence_entities:
                        continue
                    sim_index = torch.argmax(hyp_f_score)
                    nli_label = self.nli.predict([hyp_sentence], [ref_report[sim_index]])[1][0]
                    if nli_label == 'entailment':
                        match_p += 1
                    for entity in hyp_sentence_entities:
                        total_p += 1
                        if nli_label == 'contradiction':
                            continue
                        if entity in entity_ner_r:
                            match_p += 1

                match_r = 0
                entity_ner_h = list(chain.from_iterable(ner_h))
                total_r = 0
                for ref_sentence, ref_sentence_entities, ref_f_score in zip(ref_report, ner_r, f_scores.T):
                    # No entites in current sentence
                    if not ref_sentence_entities:
                        continue
                    sim_index = torch.argmax(ref_f_score)
                    nli_label = self.nli.predict([ref_sentence], [hyp_report[sim_index]])[1][0]
                    if nli_label == 'entailment':
                        match_r += 1
                    for entity in ref_sentence_entities:
                        total_r += 1
                        if nli_label == 'contradiction':
                            continue
                        if entity in entity_ner_h:
                            match_r += 1

                pr_e = match_p / total_p if total_p > 0 else 0.0
                rc_e = match_r / total_r if total_r > 0 else 0.0

                # harmonic mean
                score_e = 2 * pr_e * rc_e / (pr_e + rc_e) if pr_e > 0.0 and rc_e > 0.0 else 0.0
                scores_e.append(score_e)

            mean_exact_e = np.mean(scores_e)
            return mean_exact_e, scores_e


if __name__ == '__main__':
    num = str(776414)

    l1 = open("test_best-1_" + num + "_hyps.txt").readlines()
    # l1 = [l.strip() for l in l1][:10]
    l1 = [l.strip() for l in l1]
    l2 = open("test_best-1_" + num + "_refs.txt").readlines()
    # l2 = [l.strip() for l in l2][:10]
    l2 = [l.strip() for l in l2]

    # x = RadEntityNLI()(
    #     refs=[
    #         'no evidence of consolidation to suggest pneumonia is seen. there  is some retrocardiac atelectasis. a small left pleural effusion may be  present. no pneumothorax is seen. no pulmonary edema. a right granuloma is  unchanged. the heart is mildly enlarged, unchanged. there is tortuosity of  the aorta.',
    #         'there are moderate bilateral pleural effusions with overlying atelectasis,  underlying consolidation not excluded. mild prominence of the interstitial  markings suggests mild pulmonary edema. the cardiac silhouette is mildly  enlarged. the mediastinal contours are unremarkable. there is no evidence of  pneumothorax.'
    #     ],
    #     hyps=[
    #         'heart size is moderately enlarged. the mediastinal and hilar contours are unchanged. there is no pulmonary edema. small left pleural effusion is present. patchy opacities in the lung bases likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.',
    #         'heart size is mildly enlarged. the mediastinal and hilar contours are normal. there is mild pulmonary edema. moderate bilateral pleural effusions are present, left greater than right. bibasilar airspace opacities likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.'
    #     ])

    x = RadEntityNLI()(
        refs=l2,
        hyps=l1)

    # 0.3200814901900191,
    print(x[0])
    # (0.5238658777120316, [0.5743589743589744, 0.4733727810650888])



import torch
from torch.utils.data import Dataset

from .base.ImageDataset import ImageDataset
from .base.LabelDataset import LabelDataset


class ImLabel(Dataset):
    def __init__(self, label, image, split, ckpt_dir, **kwargs):
        self.split = split
        self.image = ImageDataset(**image, split=split)
        self.label = LabelDataset(**label, split=split, ckpt_dir=ckpt_dir)

        assert len(self.image) == len(self.label)

        self.labels_map = self.label.labels_map

    def __getitem__(self, index):
        return {**self.image.__getitem__(index), **self.label.__getitem__(index)}

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {**self.image.get_collate_fn()(batch),
                        **self.label.get_collate_fn()(batch)}
            return collated

        return collate_fn

    def inference(self, image=None, label=None):
        if label is None and image is None:
            return dict()

        batch = {}
        if image is not None:
            batch.update(self.image.inference(image))

        if label is not None:
            batch.update(self.label.inference(label))

        assert len(set([len(v) for k, v in batch.items()])) == 1, 'element in batch dont have the same size'
        return batch

    def __len__(self):
        return len(self.image)

    def __repr__(self):
        return "ImLabel\n" + str(self.image) + '\n' + str(self.label)



import torch
import torch.nn as nn
from vilmedic.models.utils import get_n_params

from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.decoder.decoder_model import DecoderModel
from vilmedic.blocks.huggingface.decoder.evaluation import evaluation
from einops import rearrange


class RRG_2D_CT(nn.Module):

    def __init__(self, decoder, cnn, dl=None, **kwargs):
        super().__init__()
        # Decoder
        if dl:
            decoder.vocab_size = dl.dataset.seq.tokenizer.vocab_size
        self.dec = DecoderModel(decoder)

        # Encoder
        visual_embedding_dim = cnn.pop("visual_embedding_dim", None)
        cnn = eval(cnn.pop('proto'))(**cnn)
        if visual_embedding_dim:
            visual_projection = nn.Linear(visual_embedding_dim, self.dec.decoder.config.hidden_size)
        else:
            visual_projection = nn.Identity()

        self.enc = nn.Sequential(cnn, visual_projection)

        # Evaluation
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, images_mask=None, encoder_outputs=None,
                encoder_attention_mask=None, epoch=None, iteration=None, **kwargs):

        if torch.cuda.is_available():
            input_ids = input_ids.cuda()
            attention_mask = attention_mask.cuda()

        if encoder_outputs is None:
            encoder_outputs, encoder_attention_mask = self.encode(images, images_mask, **kwargs)

        out = self.dec(input_ids=input_ids,
                       attention_mask=attention_mask,
                       encoder_outputs=encoder_outputs,
                       encoder_attention_mask=encoder_attention_mask,
                       **kwargs)

        return out

    # Necessary for generation
    def encode(self, images, images_mask=None, **kwargs):
        if torch.cuda.is_available():
            images = images.cuda()

        outputs = []
        for i in range(images.size(-1)):
            slice_ = images[..., i]
            output = self.enc(slice_)
            outputs.append(output)

        features = torch.stack(outputs, dim=1)
        feature_mask = (torch.sum(torch.abs(features), dim=-1) != 0)
        return features, feature_mask

    def __repr__(self):
        s = "model: RRG_2D_CT\n"
        s += "(enc):" + str(self.enc) + '\n'
        s += "(dec):" + str(self.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



# -*- coding: utf-8 -*-
import torch
from torch import nn


# Libovický, J., & Helcl, J. (2017). Attention Strategies for Multi-Source
# Sequence-to-Sequence Learning. In Proceedings of the 55th Annual Meeting of
# the Association for Computational Linguistics (Volume 2: Short Papers)
# (Vol. 2, pp. 196-202). [Code contributed by @jlibovicky]


class HierarchicalAttention(nn.Module):
    """Hierarchical attention over multiple modalities."""

    def __init__(self, ctx_dims, hid_dim, mid_dim, att_activ='tanh'):
        super().__init__()

        self.activ = getattr(torch, att_activ)
        self.ctx_dims = ctx_dims
        self.hid_dim = hid_dim
        self.mid_dim = mid_dim

        self.ctx_projs = nn.ModuleList([
            nn.Linear(dim, mid_dim, bias=False) for dim in self.ctx_dims])
        self.dec_proj = nn.Linear(hid_dim, mid_dim, bias=True)
        self.mlp = nn.Linear(self.mid_dim, 1, bias=False)

    def forward(self, contexts, hid):
        dec_state_proj = self.dec_proj(hid)
        ctx_projected = torch.cat([
            p(ctx).unsqueeze(0) for p, ctx
            in zip(self.ctx_projs, contexts)], dim=0)
        energies = self.mlp(self.activ(dec_state_proj + ctx_projected))
        att_dist = torch.nn.functional.softmax(energies, dim=0)

        ctxs_cat = torch.cat([c.unsqueeze(0) for c in contexts])
        joint_context = (att_dist * ctxs_cat).sum(0)

        return att_dist, joint_context



import tqdm
import os
import logging
from torchvision import transforms
import copy
import matplotlib.pyplot as plt
import numpy as np
import skimage.transform
import matplotlib.cm as cm
import math
from sklearn.manifold import TSNE
import matplotlib.pyplot as plt
import umap
import omegaconf
import seaborn as sns

sns.set_theme(style="darkgrid")



def plot_attention(results, pp_dir, seed, logger, split, epoch, dl, smooth=True, **kwargs):
    if 'attentions' not in results:
        logger.warn("No attention weights found in results, skipping")
        return
    # Attention are of size num_models x bs x num_layers x num_heads x seq_len x seq_len
    attentions = results["attentions"]

    # Some checks
    dataset = copy.deepcopy(dl.dataset)
    if not len(dataset) == len(attentions):
        logger.critical("Attention: This is weird")
        return
    if not hasattr(dataset, 'transform'):
        logger.error("Attention: Dataset does not have a transform attribute")
        return

    out_dir = os.path.join(pp_dir, "plot_attention_{}_{}_{}".format(seed, split, epoch))
    os.makedirs(out_dir, exist_ok=True)

    # Getting rid of the normalization function if any
    new_transform = [t for t in dataset.transform.transforms if 'Normalize' not in t.__class__.__name__]
    dataset.transform.transforms = new_transform

    for sample in dataset:
        image = sample["image"]
        im = transforms.ToPILImage()(image)
        width, height = im.size
        if not width == height:
            logger.error("Attention: image width and height are different")
            return

        plt.imshow(im)
        weights = attentions[-1, -1, -1, -1, -1, :]
        square = int(math.sqrt(len(weights)))

        if smooth:
            alpha_im = skimage.transform.pyramid_expand(attentions[-1, -1, -1, -1, -1, :].reshape(square, square),
                                                        upscale=int(width / square),
                                                        sigma=20)
        else:
            alpha_im = skimage.transform.resize(attentions[-1, -1, -1, -1, -1, :].reshape(square, square),
                                                [width, height])
        plt.imshow(alpha_im, alpha=0.8)
        plt.set_cmap(cm.Reds)
        # plt.clim(0, 1)
        plt.colorbar()
        plt.axis('off')
        plt.savefig(os.path.join(out_dir, "att"))


def save_representation(keys, results, pp_dir, seed, logger, split, **kwargs):
    out_dir = os.path.join(pp_dir, "save_representation_{}_{}".format(seed, split))
    os.makedirs(out_dir, exist_ok=True)
    for key in keys:
        if key not in results:
            logger.warn("Key {} is not found in results dictionary")
            continue

        embeddings = np.array([vector.numpy() for vector in results[key]])
        results.pop(key)
        np.save(os.path.join(out_dir, split
                             + '_'
                             + str(key)
                             + '_'
                             + "embeddings"), embeddings)


def plot_representation(keys, results, pp_dir, seed, logger, split, dl, labels_keep=None, max_samples_per_class=None,
                        **kwargs):
    # Getting labels
    attr = [k for k, v in dl.dataset.__dict__.items() if 'LabelDataset' in str(v)]
    label_dataset = getattr(dl.dataset, attr[0])
    labels = label_dataset.labels
    labels_map = label_dataset.labels_map.idx2label
    multi_label = label_dataset.labels_map.multi_label
    out_dir = os.path.join(pp_dir, "plot_representation_{}_{}".format(seed, split))
    os.makedirs(out_dir, exist_ok=True)

    # computing repr
    for key in keys:
        embeddings, emb_labels = list(), list()
        if key not in results:
            logger.warn("Key {} is not found in results dictionary")
            continue

        for vector, label in zip(results[key], labels):
            if multi_label:
                c = np.where(label == 1.)[0]
                if labels_keep is not None:
                    c = [c_ for c_ in c if labels_map[c_] in labels_keep]
                if len(c) != 1:
                    continue  # cant plot a point that belongs to more than one class (or no class from labels_keep)
                label = c[0]

            else:
                if labels_keep is not None:
                    if not labels_map[label] in labels_keep:
                        continue

            emb_labels.append(labels_map[label])
            embeddings.append(vector.numpy())

        emb_labels = np.array(emb_labels)
        embeddings = np.array(embeddings)

        np.save(os.path.join(out_dir, split
                             + '_'
                             + str(key)
                             + '_'
                             + "embeddings"), embeddings)

        np.save(os.path.join(out_dir, split
                             + '_'
                             + str(key)
                             + '_'
                             + "labels"), emb_labels)

        assert len(embeddings) != 0, logging.error("No embedding kept for visualization")

        # Filtering the number of samples per class
        if max_samples_per_class is not None:
            if not isinstance(max_samples_per_class, int):
                logger.warn("Argument max_samples_per_class is not an integer, found {}. Using all points".format(
                    type(max_samples_per_class)))
            else:
                new_labels = []
                new_embeddings = []
                for g in np.unique(emb_labels):
                    ix = np.where(emb_labels == g)[0]
                    np.random.shuffle(ix)
                    new_labels.append(emb_labels[ix[:250]])
                    new_embeddings.append(embeddings[ix[:250]])

                embeddings = np.concatenate(new_embeddings)
                emb_labels = np.concatenate(new_labels)

        # Compute and save plots
        for visualization in [TSNE(n_components=2, n_jobs=4, verbose=0, n_iter=2000),
                              umap.UMAP(n_neighbors=len(labels_map)),
                              ]:

            visualization_name = type(visualization).__name__
            logger.settings('Computing embeddings using {}'.format(visualization_name))

            embeddings_x = visualization.fit_transform(embeddings)

            # Plotting
            fig = plt.figure()
            for g in np.unique(emb_labels):
                ix = np.where(emb_labels == g)
                plt.scatter(embeddings_x[ix, 0], embeddings_x[ix, 1], s=0.1,
                            cmap='Spectral', label=g)

            plt.legend(markerscale=10, loc='center left', bbox_to_anchor=(1, 0.5))
            plt.tight_layout()
            filename = os.path.join(out_dir, split
                                    + '_'
                                    + str(key)
                                    + '_'
                                    + visualization_name
                                    + '.png')
            fig.savefig(filename)
            plt.close()
            logger.settings('Saved as {}'.format(filename))
    return


def post_processing(post_processing, results, ckpt_dir, seed, dl, **kwargs):
    if post_processing is None:
        return

    logger = logging.getLogger(str(seed))
    pp_dir = os.path.join(ckpt_dir, 'post_processing')
    os.makedirs(pp_dir, exist_ok=True)
    for pp in post_processing:
        if "plot_attention" in pp:  # Plot attention weights
            plot_attention(results=results, pp_dir=pp_dir, seed=seed, logger=logger, **kwargs)
        if "plot_representation" in pp:
            plot_representation(results=results, pp_dir=pp_dir, seed=seed, logger=logger, dl=dl,
                                **pp["plot_representation"], **kwargs)
        if "save_representation" in pp:
            save_representation(results=results, pp_dir=pp_dir, seed=seed, logger=logger,
                                **pp["save_representation"], **kwargs)
        else:
            logger.warn("Post-processing: No function implemented for '{}'".format(pp))



# Filename: cider.py
#
# Description: Describes the class to compute the CIDEr (Consensus-Based Image Description Evaluation) Metric 
#               by Vedantam, Zitnick, and Parikh (http://arxiv.org/abs/1411.5726)
#
# Creation Date: Sun Feb  8 14:16:54 2015
#
# Authors: Ramakrishna Vedantam <vrama91@vt.edu> and Tsung-Yi Lin <tl483@cornell.edu>

from .ciderD_scorer import CiderScorer
import pdb
import torch.nn as nn


class CiderD(nn.Module):
    """
    Main Class to compute the CIDEr metric 

    """

    def __init__(self, test=None, refs=None, n=4, sigma=6.0, **kwargs):
        # set cider to sum over 1 to 4-grams
        super().__init__()
        self._n = n
        # set the standard deviation parameter for gaussian penalty
        self._sigma = sigma

    def forward(self, gts, res):
        return self.compute_score(gts, res)

    def compute_score(self, gts, res):
        """
        Main function to compute CIDEr score
        :param  hypo_for_image (dict) : dictionary with key <image> and value <tokenized hypothesis / candidate sentence>
                ref_for_image (dict)  : dictionary with key <image> and value <tokenized reference sentence>
        :return: cider (float) : computed CIDEr score for the corpus 
        """

        cider_scorer = CiderScorer(n=self._n, sigma=self._sigma)
        res = {i: [v] for i, v in enumerate(res)}
        gts = {i: [v] for i, v in enumerate(gts)}
        for id in sorted(gts.keys()):
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert (type(hypo) is list)
            assert (len(hypo) == 1)
            assert (type(ref) is list)
            assert (len(ref) > 0)

            cider_scorer += (hypo[0], ref)

        (score, scores) = cider_scorer.compute_score()

        return score, scores

    def method(self):
        return "CIDEr"



#!/usr/bin/env python

# Python wrapper for METEOR implementation, by Xinlei Chen
# Acknowledge Michael Denkowski for the generous discussion and help
from __future__ import division

import atexit
import logging
import os
import re
import subprocess
import sys
import threading

import psutil
import torch.nn as nn

# Assumes meteor-1.5.jar is in the same directory as meteor.py.  Change as needed.
METEOR_JAR = 'meteor-1.5.jar'


def enc(s):
    return s.encode('utf-8')


def dec(s):
    return s.decode('utf-8')


class Meteor(nn.Module):

    def __init__(self, **kwargs):
        # Used to guarantee thread safety
        super().__init__()
        self.lock = threading.Lock()

        mem = '2G'
        mem_available_G = psutil.virtual_memory().available / 1E9
        if mem_available_G < 2:
            logging.warning("There is less than 2GB of available memory.\n"
                            "Will try with limiting Meteor to 1GB of memory but this might cause issues.\n"
                            "If you have problems using Meteor, "
                            "then you can try to lower the `mem` variable in meteor.py")
            mem = '1G'

        meteor_cmd = ['java', '-jar', '-Xmx{}'.format(mem), METEOR_JAR,
                      '-', '-', '-stdio', '-l', 'en', '-norm']
        env = os.environ.copy()
        env['LC_ALL'] = "C"
        self.meteor_p = subprocess.Popen(meteor_cmd,
                                         cwd=os.path.dirname(os.path.abspath(__file__)),
                                         env=env,
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)

        atexit.register(self.close)

    def close(self):
        with self.lock:
            if self.meteor_p:
                self.meteor_p.kill()
                self.meteor_p.wait()
                self.meteor_p = None
        # if the user calls close() manually, remove the
        # reference from atexit so the object can be garbage-collected.
        if atexit is not None and atexit.unregister is not None:
            atexit.unregister(self.close)

    def forward(self, gts, res):
        return self.compute_score(gts, res)

    def compute_score(self, gts, res):
        res = {i: [v] for i, v in enumerate(res)}
        gts = {i: [v] for i, v in enumerate(gts)}
        scores = []

        eval_line = 'EVAL'
        with self.lock:
            for i in sorted(gts.keys()):
                assert (len(res[i]) == 1)
                stat = self._stat(res[i][0], gts[i])
                eval_line += ' ||| {}'.format(stat)

            self.meteor_p.stdin.write(enc('{}\n'.format(eval_line)))
            self.meteor_p.stdin.flush()
            for i in range(0, len(list(gts.keys()))):
                v = self.meteor_p.stdout.readline()
                try:
                    scores.append(float(dec(v.strip())))
                except:
                    sys.stderr.write("Error handling value: {}\n".format(v))
                    sys.stderr.write("Decoded value: {}\n".format(dec(v.strip())))
                    sys.stderr.write("eval_line: {}\n".format(eval_line))
                    # You can try uncommenting the next code line to show stderr from the Meteor JAR.
                    # If the Meteor JAR is not writing to stderr, then the line will just hang.
                    # sys.stderr.write("Error from Meteor:\n{}".format(self.meteor_p.stderr.read()))
                    raise
            score = float(dec(self.meteor_p.stdout.readline()).strip())

        return score, scores

    def method(self):
        return "METEOR"

    def _stat(self, hypothesis_str, reference_list):
        # SCORE ||| reference 1 words ||| reference n words ||| hypothesis words
        hypothesis_str = hypothesis_str.replace('|||', '')
        score_line = ' ||| '.join(('SCORE', ' ||| '.join(reference_list), hypothesis_str))
        score_line = re.sub(r'\s+', ' ', score_line)
        self.meteor_p.stdin.write(enc(score_line))
        self.meteor_p.stdin.write(enc('\n'))
        self.meteor_p.stdin.flush()
        return dec(self.meteor_p.stdout.readline()).strip()

    def _score(self, hypothesis_str, reference_list):
        with self.lock:
            # SCORE ||| reference 1 words ||| reference n words ||| hypothesis words
            hypothesis_str = hypothesis_str.replace('|||', '').replace('  ', ' ')
            score_line = ' ||| '.join(('SCORE', ' ||| '.join(reference_list), hypothesis_str))
            self.meteor_p.stdin.write(enc('{}\n'.format(score_line)))
            self.meteor_p.stdin.flush()
            stats = dec(self.meteor_p.stdout.readline()).strip()
            eval_line = 'EVAL ||| {}'.format(stats)
            # EVAL ||| stats 
            self.meteor_p.stdin.write(enc('{}\n'.format(eval_line)))
            self.meteor_p.stdin.flush()
            score = float(dec(self.meteor_p.stdout.readline()).strip())
            # bug fix: there are two values returned by the jar file, one average, and one all, so do it twice
            # thanks for Andrej for pointing this out
            score = float(dec(self.meteor_p.stdout.readline()).strip())
        return score

    def __del__(self):
        self.close()



import torch
import torch.nn as nn
from vilmedic.models.utils import get_n_params

from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.encoder.encoder_model import EncoderModel
from vilmedic.blocks.losses import GLoRIALoss, gloria_attention_fn, cosine_similarity

from tqdm import tqdm
import numpy as np
from sklearn import metrics


def evaluation(models, config, dl, from_training, **kwargs):
    # No ensembling for this evaluation
    model = models[0]

    losses = []
    linguistics = []
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for i, batch in enumerate(pbar):
        batch = {k: v.cuda() if isinstance(v, torch.Tensor) else v for k, v in batch.items()}
        out = model(**batch)
        losses.append(out['loss'].mean().cpu().data.numpy())

        if not from_training:
            linguistics.append(out['sent_embeddings'].cpu().data)
            visuals.append(out['global_features'].cpu().data)

    if from_training:
        return {'loss': np.ndarray.mean(np.array(losses))}

    return {'loss': np.ndarray.mean(np.array(losses)),
            'linguistic': torch.cat(linguistics),
            'visual': torch.cat(visuals)
            }


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


class GLoRIA(nn.Module):

    def __init__(self, encoder, cnn, visual_embedder, loss, dl, forward_batch_size=12, **kwargs):
        super().__init__()

        # Linguistic encoder
        self.linguistic = EncoderModel(encoder)
        self.last_n_layers = encoder.last_n_layers
        self.idxtoword = {v: k for k, v in dl.dataset.tokenizer.get_vocab().items()}

        # Visual Encoder
        self.visual = eval(cnn.pop('proto'))(**cnn)
        self.global_embedder = nn.Linear(visual_embedder.feature_dim, self.linguistic.encoder.config.hidden_size)
        self.local_embedder = nn.Conv2d(visual_embedder.interm_feature_dim,
                                        self.linguistic.encoder.config.hidden_size,
                                        kernel_size=1,
                                        stride=1,
                                        padding=0,
                                        bias=False)
        self.up_sample = nn.Upsample(size=(299, 299), mode="bilinear", align_corners=True)
        # Loss
        self.loss_fn = GLoRIALoss(**loss)

        # Hook
        self.activation = {}

        def getActivation(name):
            def hook(model, input, output):
                self.activation[name] = output

            return hook

        self.visual.cnn[6].register_forward_hook(getActivation('local_features'))  # corresponds to the ouput of layer3

        # Evaluation
        self.eval_func = evaluation
        self.fbs = forward_batch_size

    def forward(self, input_ids, attention_mask, images, **kwargs):
        bs = images.shape[0]
        global_features = []
        local_features = []
        hidden_states = []

        # Forward passes
        for i in list(chunks(range(bs), min(self.fbs, bs))):
            input_ids_ = input_ids[i]
            attention_mask_ = attention_mask[i]
            images_ = images[i]

            global_features.append(self.global_embedder(self.visual(self.up_sample(images_.cuda()))))
            local_features.append(self.local_embedder(self.activation["local_features"]))
            output = self.linguistic(input_ids_.cuda(),
                                     attention_mask_.cuda(),
                                     output_hidden_states=True)
            hidden_states.append(torch.stack(output["hidden_states"]))

        global_features = torch.cat(global_features)
        local_features = torch.cat(local_features)
        hidden_states = torch.cat(hidden_states, dim=1)

        # Aggregate Token embeddings
        embeddings = hidden_states[-self.last_n_layers:]
        embeddings, sents = self.aggregate_tokens(embeddings, input_ids)

        # Linguistic local and global embeddings
        sent_embeddings = torch.sum(torch.mean(embeddings, dim=2), dim=1)
        word_embeddings = torch.sum(embeddings, dim=1)
        word_embeddings = word_embeddings.permute(0, 2, 1)

        # Losses
        loss, attention_maps = self.loss_fn(global_features, local_features, word_embeddings, sent_embeddings, sents)

        return {"loss": loss, "global_features": global_features, "local_features": local_features,
                "word_embeddings": word_embeddings, "sent_embeddings": sent_embeddings}

    def aggregate_tokens(self, embeddings, input_ids):
        num_layers, batch_size, num_words, dim = embeddings.shape
        embeddings = embeddings.permute(1, 2, 0, 3)

        agg_embs_batch = []
        sentences = []

        # loop over batch
        for embs, caption_id in zip(embeddings, input_ids):

            agg_embs = []
            token_bank = []
            words = []
            word_bank = []

            # loop over sentence
            for word_emb, word_id in zip(embs, caption_id):

                word = self.idxtoword[word_id.item()]

                if word == "[SEP]":
                    new_emb = torch.stack(token_bank)
                    new_emb = new_emb.sum(axis=0)
                    agg_embs.append(new_emb)
                    words.append("".join(word_bank))

                    agg_embs.append(word_emb)
                    words.append(word)
                    break

                if not word.startswith("##"):
                    if len(word_bank) == 0:
                        token_bank.append(word_emb)
                        word_bank.append(word)
                    else:
                        new_emb = torch.stack(token_bank)
                        new_emb = new_emb.sum(axis=0)
                        agg_embs.append(new_emb)
                        words.append("".join(word_bank))

                        token_bank = [word_emb]
                        word_bank = [word]
                else:
                    if word.startswith("##"):
                        token_bank.append(word_emb)
                        word_bank.append(word[2:])

            agg_embs = torch.stack(agg_embs)
            padding_size = num_words - len(agg_embs)
            paddings = torch.zeros(padding_size, num_layers, dim)
            paddings = paddings.to(agg_embs.device)
            words = words + ["[PAD]"] * padding_size

            agg_embs_batch.append(torch.cat([agg_embs, paddings]))
            sentences.append(words)

        agg_embs_batch = torch.stack(agg_embs_batch)
        agg_embs_batch = agg_embs_batch.permute(0, 2, 1, 3)
        return agg_embs_batch, sentences

    def zero_shot_classification(self, input_ids, attention_mask, images):

        # get similarities for each class
        similarities = self.get_similarities(
            input_ids, attention_mask, images, similarity_type="both"
        )
        class_similarities = similarities.max(axis=1)  # average between class prompts
        return class_similarities

    def get_similarities(self, input_ids, attention_mask, images, similarity_type="both"):

        # warnings
        if similarity_type not in ["global", "local", "both"]:
            raise RuntimeError(
                f"similarity type should be one of ['global', 'local', 'both']"
            )

        cap_lens = [len([t for t in i if t != 0]) for i in input_ids]

        # get global and local image features
        with torch.no_grad():
            out = self(input_ids, attention_mask, images)

        img_emb_g, img_emb_l, text_emb_l, text_emb_g = out["global_features"], out["local_features"], out[
            "word_embeddings"], out["sent_embeddings"]

        # get similarities
        global_similarities = self.get_global_similarities(img_emb_g, text_emb_g)
        local_similarities = self.get_local_similarities(
            img_emb_l, text_emb_l, cap_lens
        )
        similarities = (local_similarities + global_similarities) / 2

        if similarity_type == "global":
            return global_similarities.detach().cpu().numpy()
        elif similarity_type == "local":
            return local_similarities.detach().cpu().numpy()
        else:
            return similarities.detach().cpu().numpy()

    def get_global_similarities(self, img_emb_g, text_emb_g):
        img_emb_g = img_emb_g.detach().cpu().numpy()
        text_emb_g = text_emb_g.detach().cpu().numpy()
        global_similarities = metrics.pairwise.cosine_similarity(img_emb_g, text_emb_g)
        global_similarities = torch.tensor(global_similarities)
        return global_similarities

    def get_local_similarities(self, img_emb_l, text_emb_l, cap_lens):

        batch_size = img_emb_l.shape[0]
        similarities = []

        for i in range(len(text_emb_l)):
            words_num = cap_lens[i]
            word = (
                text_emb_l[i, :, 1: words_num + 1].unsqueeze(0).contiguous()
            )  # [1, 768, 25]

            word = word.repeat(batch_size, 1, 1)  # [48, 768, 25]
            context = img_emb_l  # [48, 768, 19, 19]

            weiContext, attn = gloria_attention_fn(
                word, context, 4.0
            )  # [48, 768, 25], [48, 25, 19, 19]

            word = word.transpose(1, 2).contiguous()  # [48, 25, 768]
            weiContext = weiContext.transpose(1, 2).contiguous()  # [48, 25, 768]

            word = word.view(batch_size * words_num, -1)  # [1200, 768]
            weiContext = weiContext.view(batch_size * words_num, -1)  # [1200, 768]
            #
            row_sim = cosine_similarity(word, weiContext)
            row_sim = row_sim.view(batch_size, words_num)  # [48, 25]

            row_sim.mul_(5.0).exp_()
            row_sim, max_row_idx = torch.max(row_sim, dim=1, keepdim=True)  # [48, 1]

            row_sim = torch.log(row_sim)

            similarities.append(row_sim)

        local_similarities = torch.cat(similarities, 1).detach().cpu()

        return local_similarities

    def __repr__(self):
        s = super().__repr__() + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch.nn as nn
from vilmedic.models.utils import get_n_params
import functools


from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.decoder.evaluation import evaluation as evaluation_
from vilmedic.blocks.scorers.scores import CheXbert
from .RRG import RRG
import numpy as np
import copy
import torch
from transformers import top_k_top_p_filtering
import torch.nn.functional as F
from torch.nn import Identity
from .PPOTrainer import PPOTrainer, stack_dicts
import random


def evaluation(models, config, dl, **kwargs):
    models = [m.model for m in models]  # Get trained RRG instance
    return evaluation_(models, config, dl)


class ValueHead(nn.Module):
    """The ValueHead class implements a head for GPT2 that returns a scalar for each output token."""

    def __init__(self):
        super().__init__()
        self.detach_head = False
        self.summary = nn.Linear(768, 1)
        self.activation = Identity()
        self.first_dropout = nn.Dropout(0.1)
        self.last_dropout = Identity()
        self.flatten = nn.Flatten()

    def forward(self, hidden_states, cls_index=None):
        output = hidden_states
        output = self.first_dropout(output)
        output = self.summary(output)
        output = self.activation(output)
        output = self.last_dropout(output)

        return output


class RRG_PPO(nn.Module):

    def __init__(self, decoder, cnn, ckpt, ppo=None, **kwargs):
        super().__init__()

        # Models
        state_dict = torch.load(ckpt)["model"]
        self.model = RRG(copy.deepcopy(decoder), copy.deepcopy(cnn), **kwargs)
        self.model.load_state_dict(state_dict, strict=True)
        self.model_ref = RRG(copy.deepcopy(decoder), copy.deepcopy(cnn), **kwargs)
        self.model_ref.load_state_dict(state_dict, strict=True)

        self.v_head = ValueHead()
        del state_dict

        # PPO
        if ppo is None:
            ppo = {}
        self.ppo = PPOTrainer(self.model.encode, self.model.dec, self.model_ref.dec, self.v_head, **ppo)

        # Scoring
        self.chexbert = CheXbert()

        # Tokenizers
        self.bos_token_id = self.model.dec.decoder.config.bos_token_id
        self.eos_token_id = self.model.dec.decoder.config.eos_token_id

        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, encoder_outputs=None, **kwargs):
        # Tokenizer
        assert 'dl' in kwargs
        dataset = kwargs["dl"].dataset
        tokenizer = dataset.tokenizer

        with torch.no_grad():
            # 0. get encoder states
            encoder_hidden_states = self.model.encode(images.cuda())

            # 1. Rollout
            batch_size = input_ids.shape[0]
            seq_len = input_ids.shape[1]
            rollout_input_ids = torch.ones((batch_size, 1), dtype=torch.long).cuda() * self.bos_token_id
            for i in range(seq_len):
                outputs = self.model.dec(
                    rollout_input_ids,
                    attention_mask=rollout_input_ids.new_ones(rollout_input_ids.shape, dtype=torch.long),
                    encoder_outputs=encoder_hidden_states
                )

                next_token_logits = outputs["logits"][:, -1, :]
                next_token_logits = top_k_top_p_filtering(next_token_logits,
                                                          top_k=self.ppo.ppo_params["top_k"],
                                                          top_p=self.ppo.ppo_params["top_p"])
                # Sample
                probs = F.softmax(next_token_logits, dim=-1)
                next_token = torch.multinomial(probs, num_samples=1).squeeze(1)
                rollout_input_ids = torch.cat([rollout_input_ids, next_token.unsqueeze(-1)], dim=-1)

            # 2. Get reward
            hyp_list = []
            ref_list = []
            for h, r in zip(rollout_input_ids, input_ids):
                hyp_list.append(tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
                ref_list.append(tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))

            chexbert, _ = self.chexbert(hyp_list, ref_list)
            scores = torch.tensor([chexbert["micro avg"]["f1-score"] for _ in range(len(hyp_list))])
            rollout_input_ids = tokenizer(hyp_list, **dataset.tokenizer_args)

            # 3. Optimization
            logprobs, ref_logprobs, values = self.ppo.batched_forward_pass(rollout_input_ids.input_ids.cuda(),
                                                                           rollout_input_ids.attention_mask.cuda(),
                                                                           encoder_hidden_states)

            rewards, non_score_reward, kl_coef = self.ppo.compute_rewards(scores, logprobs, ref_logprobs)

        all_stats = []
        idxs = list(range(batch_size))
        for _ in range(self.ppo.ppo_params['ppo_epochs']):
            random.shuffle(idxs)
            for i in range(batch_size):
                idx = idxs[i]
                train_stats = self.ppo.train_minibatch(seq_len,
                                                       logprobs[idx:idx + 1],
                                                       values[idx:idx + 1],
                                                       rewards[idx:idx + 1],
                                                       rollout_input_ids.input_ids[idx:idx + 1].cuda(),
                                                       rollout_input_ids.attention_mask[idx:idx + 1].cuda(),
                                                       images[idx:idx + 1].cuda())
                all_stats.append(train_stats)

        # Handle stats
        train_stats = stack_dicts(all_stats)
        train_stats['policy/advantages'] = torch.flatten(train_stats['policy/advantages']).unsqueeze(0)
        train_stats['policy/ratio'] = torch.flatten(train_stats['policy/ratio']).unsqueeze(0)
        stats = self.ppo.get_stats(scores=scores,
                                   logprobs=logprobs,
                                   ref_logprobs=ref_logprobs,
                                   non_score_reward=non_score_reward,
                                   train_stats=train_stats,
                                   kl_coef=kl_coef)

        self.ppo.kl_ctl.update(stats['objective/kl'][-1], batch_size)

        return "objective/kl: {}, objective/kl_coef: {},  objective/score: {}, ppo/mean_non_score_reward: {}".format(
            np.mean(stats['objective/kl']),
            np.mean(stats['objective/kl_coef']),
            np.mean(stats['objective/score']),
            np.mean(stats['ppo/mean_non_score_reward'])
        )

    def __repr__(self):
        s = "RRG_PPO\n"
        # s += str(self.enc) + '\n'
        # s += str(self.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
import torch.nn as nn


class ConVIRTLoss(nn.Module):
    def __init__(self, tau, lambda_, **kwargs):
        super().__init__()
        self.tau = tau
        self.lambda_ = lambda_
        self.cos_loss = nn.CosineSimilarity()

    def forward(self, linguistic, visual):
        nominator = torch.exp(torch.div(self.cos_loss(linguistic, visual), self.tau))

        denominator_l = self.pairwise_cosine_distance(linguistic, visual)
        denominator_l = torch.sum(torch.exp(torch.div(denominator_l, self.tau)), dim=1)
        loss_l = -torch.log(torch.div(nominator, denominator_l))

        denominator_v = self.pairwise_cosine_distance(visual, linguistic)
        denominator_v = torch.sum(torch.exp(torch.div(denominator_v, self.tau)), dim=1)
        loss_v = -torch.log(torch.div(nominator, denominator_v))
        loss = torch.mean(self.lambda_ * loss_v + (1 - self.lambda_) * loss_l)
        return loss, loss_l, loss_v

    @staticmethod
    def pairwise_cosine_distance(a, b):
        eps = 1e-08
        a_n, b_n = a.norm(dim=1)[:, None], b.norm(dim=1)[:, None]
        a_norm = a / torch.clamp(a_n, min=eps)
        b_norm = b / torch.clamp(b_n, min=eps)
        return torch.mm(a_norm, b_norm.transpose(0, 1))

    def __repr__(self):
        return "ConVIRTLoss(\n" + \
               "\t(cos_loss): CosineSimilarity()\n" + \
               "\t(tau): {}\n".format(self.tau) + \
               "\t(lambda_): {}\n".format(self.lambda_) + \
               ")"



import torch
from torch import nn
import torch.nn.functional as F
from einops import repeat

from vit_pytorch.vit import Transformer

class MAE(nn.Module):
    def __init__(
        self,
        *,
        encoder,
        decoder_dim,
        masking_ratio = 0.75,
        decoder_depth = 1,
        decoder_heads = 8,
        decoder_dim_head = 64
    ):
        super().__init__()
        assert masking_ratio > 0 and masking_ratio < 1, 'masking ratio must be kept between 0 and 1'
        self.masking_ratio = masking_ratio

        # extract some hyperparameters and functions from encoder (vision transformer to be trained)

        self.encoder = encoder
        num_patches, encoder_dim = encoder.pos_embedding.shape[-2:]
        self.to_patch, self.patch_to_emb = encoder.to_patch_embedding[:2]
        pixel_values_per_patch = self.patch_to_emb.weight.shape[-1]

        # decoder parameters

        self.enc_to_dec = nn.Linear(encoder_dim, decoder_dim) if encoder_dim != decoder_dim else nn.Identity()
        self.mask_token = nn.Parameter(torch.randn(decoder_dim))
        self.decoder = Transformer(dim = decoder_dim, depth = decoder_depth, heads = decoder_heads, dim_head = decoder_dim_head, mlp_dim = decoder_dim * 4)
        self.decoder_pos_emb = nn.Embedding(num_patches, decoder_dim)
        self.to_pixels = nn.Linear(decoder_dim, pixel_values_per_patch)

    def forward(self, img):
        resisc45_images/test/ = img.device

        # get patches

        patches = self.to_patch(img)
        batch, num_patches, *_ = patches.shape

        # patch to encoder tokens and add positions

        tokens = self.patch_to_emb(patches)
        tokens = tokens + self.encoder.pos_embedding[:, 1:(num_patches + 1)]

        # calculate of patches needed to be masked, and get random indices, dividing it up for mask vs unmasked

        num_masked = int(self.masking_ratio * num_patches)
        rand_indices = torch.rand(batch, num_patches, device = device).argsort(dim = -1)
        masked_indices, unmasked_indices = rand_indices[:, :num_masked], rand_indices[:, num_masked:]

        # get the unmasked tokens to be encoded

        batch_range = torch.arange(batch, device = device)[:, None]
        tokens = tokens[batch_range, unmasked_indices]

        # get the patches to be masked for the final reconstruction loss

        masked_patches = patches[batch_range, masked_indices]

        # attend with vision transformer

        encoded_tokens = self.encoder.transformer(tokens)

        # project encoder to decoder dimensions, if they are not equal - the paper says you can get away with a smaller dimension for decoder

        decoder_tokens = self.enc_to_dec(encoded_tokens)

        # reapply decoder position embedding to unmasked tokens

        decoder_tokens = decoder_tokens + self.decoder_pos_emb(unmasked_indices)

        # repeat mask tokens for number of masked, and add the positions using the masked indices derived above

        mask_tokens = repeat(self.mask_token, 'd -> b n d', b = batch, n = num_masked)
        mask_tokens = mask_tokens + self.decoder_pos_emb(masked_indices)

        # concat the masked tokens to the decoder tokens and attend with decoder

        decoder_tokens = torch.cat((mask_tokens, decoder_tokens), dim = 1)
        decoded_tokens = self.decoder(decoder_tokens)

        # splice out the mask tokens and project to pixel values

        mask_tokens = decoded_tokens[:, :num_masked]
        pred_pixel_values = self.to_pixels(mask_tokens)

        # calculate reconstruction loss

        recon_loss = F.mse_loss(pred_pixel_values, masked_patches)
        return recon_loss



import torch
import itertools
import numpy as np
from tqdm import tqdm


def evaluation(models, config, dl, **kwargs):
    logits = np.array([])
    labels = np.array([])
    losses = np.array([])
    attentions = None
    cumulative_index = 0
    post_processing = {}

    for num_batch, batch in enumerate(tqdm(dl, total=len(dl))):
        label = batch['labels']
        batch_size = label.shape[0]
        num_classes = label.shape[1]

        batch = {k: v.cuda() if (isinstance(v, torch.Tensor) and torch.cuda.is_available()) else v for k, v in
                 batch.items()}
        results = [model(**batch) for model in models]

        # Pre-allocate memory
        if num_batch == 0:
            logits = np.zeros((len(dl.dataset), len(models), num_classes))
            labels = np.zeros((len(dl.dataset), num_classes))
            losses = np.zeros((len(dl), len(models)))
            if 'attentions' in results[0]:
                # results[0]['attentions'] is num_layers x batch_size x num_heads x sequence_length x sequence_length
                num_layers = len(results[0]['attentions'])
                attention_shape = results[0]['attentions'][0][0].shape
                attentions = np.zeros((len(dl.dataset), len(models), num_layers, *attention_shape))

        # iterating over the batch, stacking refs and hyps
        for i in range(batch_size):
            for j, r in enumerate(results):
                logits[cumulative_index + i][j] = r['output'][i].data.cpu().numpy()
            labels[cumulative_index + i] = label[i].data.cpu().numpy()

        # Loss
        for j, r in enumerate(results):
            losses[num_batch][j] = r['loss'].cpu().item()

        # Handle attention weights
        if attentions is not None:
            for j, r in enumerate(results):
                for k, attention_layer in enumerate(r['attentions']):
                    for i in range(batch_size):
                        attentions[cumulative_index + i][j][k] = attention_layer[i].data.cpu().numpy()

        cumulative_index += batch_size

    preds = np.mean(logits, axis=1)
    loss = np.mean(losses)

    if attentions:
        post_processing["attentions"] = np.array(attentions)

    return {**{'loss': loss, 'refs': labels, 'hyps': preds, 'logits': logits}, **post_processing}



import torch
from vilmedic.blocks.scorers.scores import REWARD_COMPLIANT
from omegaconf.listconfig import ListConfig
import inspect
import torch.nn.functional as F
import json
import numpy as np
import torch.nn as nn


# scst loss intuition:
# https://ai.stackexchange.com/questions/2405/how-do-i-handle-negative-rewards-in-policy-gradients-with-the-cross-entropy-loss

def scst_loss(input,
              seq,
              reward_sampling,
              reward_greedy,
              scores_weights,
              pad_token_id):
    # HuggingFace TopKLogitsWarper (if top_k > 0) puts -float('inf') for non top_k (or if we use bad_words_ids)
    # Padding can then have logits -float('inf') though we have to pad because hyp generation is finished
    # masking -float('inf') * 0 results in NaN s, therefore need to do:
    input[input == -float("Inf")] = 0.
    #####

    # Masked logits
    mask = (seq > pad_token_id).float()
    input = input.squeeze(-1)
    input = input * mask
    input = input / torch.sum(mask)

    # SCST Loss
    delta_rewards = [torch.tensor(rs).cuda() - torch.tensor(rg).cuda() for rs, rg in
                     zip(reward_sampling, reward_greedy)]

    loss = [scores_weights[i] * (-input * r.unsqueeze(-1).expand_as(input)) for i, r in enumerate(delta_rewards)]
    # Compute mean loss
    #  Double sum: sum words, then sum sentences. Division is done beforehand 'input = input / torch.sum(mask)'
    loss = sum([torch.sum(l) for l in loss])

    # Stats
    delta_reward = torch.mean(torch.stack(delta_rewards))
    delta_reward_per_metric = torch.mean(torch.stack(delta_rewards), dim=-1)

    return loss, delta_reward, delta_reward_per_metric


class SCST(nn.Module):
    def __init__(self, decoder, dl, scores, scores_args=None, scores_weights=None, top_k=None, use_nll=False):
        super().__init__()

        dataset = dl.dataset
        if hasattr(dataset, "tokenizer"):
            self.tokenizer = dataset.tokenizer
            self.max_length = dataset.tokenizer_max_len
        elif hasattr(dataset, "tgt_tokenizer"):
            self.tokenizer = dataset.tgt_tokenizer
            self.max_length = dataset.tgt_tokenizer_max_len
        else:
            raise NotImplementedError("Where is tokenizer in dataset?")

        self.decoder = decoder
        self.top_k = top_k
        self.use_nll = use_nll
        self.bos_token_id = self.decoder.config.bos_token_id
        self.eos_token_id = self.decoder.config.eos_token_id
        self.pad_token_id = self.decoder.config.pad_token_id
        self.scores = scores
        self.scores_args = scores_args
        self.scores_weights = scores_weights

        assert self.scores is not None

        if not isinstance(scores, (list, ListConfig)):
            scores = [scores]

        scores = list((map(lambda x: x.lower(), scores)))
        assert all([score in REWARD_COMPLIANT for score in scores]), "{} not in {}".format(scores,
                                                                                           REWARD_COMPLIANT.keys())
        # Scores weights
        if len(scores) > 1 or use_nll:
            assert scores_weights is not None, "You need to mention scores_weights"
            assert isinstance(scores_weights, (list, ListConfig)), "scores_weights must be a list"
            if self.use_nll:
                assert len(scores_weights) == len(scores) + 1, "Mention nll_weight + as much scores_weights as scores"
            else:
                assert len(scores_weights) == len(scores), "Mention as much scores_weights as scores"

        else:
            self.scores_weights = [1.0]

        # Scores args
        if scores_args is not None:
            if not isinstance(scores_args, (list, ListConfig)):
                scores_args = [scores_args]
            assert len(scores_args) == len(scores), \
                "You need to mention as much scores_args as scores (i.e. [{arg1:value}, {}, {arg1:value}])"
        else:
            self.scores_args = [None] * len(scores)

        self.scorers = []
        self.scorers_index = []
        for index, score in enumerate(scores):
            scorer, scorer_index = REWARD_COMPLIANT[score]
            if self.scores_args[index] is not None:
                scorer = scorer(**self.scores_args[index])
            else:
                scorer = scorer()
            self.scorers.append(scorer)
            self.scorers_index.append(scorer_index)

    def forward_greedy(self, input_ids, encoder_hidden_states, encoder_attention_mask):
        assert not torch.is_grad_enabled(), "Please add torch.no_grad() decorator"
        batch_size = input_ids.shape[0]
        out = self.decoder.generate(
            input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * self.bos_token_id,
            max_length=self.max_length,
            num_beams=1,
            num_return_sequences=1,
            return_dict_in_generate=True,
            output_scores=True,
            encoder_hidden_states=encoder_hidden_states.detach(),
            encoder_attention_mask=encoder_attention_mask.detach(),
            forced_eos_token_id=True,
            use_cache=True,
        )
        greedy_input_ids = out.sequences
        reward_greedy, hyp_list, ref_list = self.get_reward(greedy_input_ids.detach().data, input_ids)
        return reward_greedy, hyp_list, ref_list

    def forward_sampling(self, input_ids, attention_mask, encoder_hidden_states, encoder_attention_mask, reward_greedy):
        assert torch.is_grad_enabled()
        batch_size = input_ids.shape[0]
        if self.use_nll:
            nll_loss = self.decoder(input_ids=input_ids.cuda(),
                                    attention_mask=attention_mask.cuda(),
                                    encoder_hidden_states=encoder_hidden_states,
                                    encoder_attention_mask=encoder_attention_mask,
                                    labels=input_ids.cuda(),
                                    )["loss"]

        out = inspect.unwrap(self.decoder.generate)(  # inspect.unwrap removes the torch.no_grad() decorator
            self=self.decoder,
            input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * self.bos_token_id,
            max_length=self.max_length,
            num_beams=1,
            num_return_sequences=1,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            bad_words_ids=[[self.pad_token_id], [self.bos_token_id]],
            top_k=self.top_k,
            forced_eos_token_id=True,
            output_scores=True,
            do_sample=True,
            use_cache=True,
            return_dict_in_generate=True,
        )
        sampled_ids = out.sequences[:, 1:].contiguous()
        logits = torch.stack(out.scores, dim=1)
        logits = F.log_softmax(logits, dim=-1)
        sampled_logits = logits.gather(2, sampled_ids.unsqueeze(-1))

        reward_sampling, hyp_list, _ = self.get_reward(sampled_ids.data, input_ids)
        loss, delta_reward, delta_reward_per_metric = scst_loss(sampled_logits,
                                                                sampled_ids.data,
                                                                reward_sampling,
                                                                reward_greedy,
                                                                # avoid nll_weight if present
                                                                self.scores_weights[-len(self.scores):],
                                                                self.pad_token_id)
        if self.use_nll:
            loss += self.scores_weights[0] * nll_loss

        return loss, delta_reward, delta_reward_per_metric, reward_sampling, hyp_list

    def get_reward(self, rollout_input_ids, input_ids):
        hyp_list = []
        ref_list = []
        for h, r in zip(rollout_input_ids, input_ids):
            hyp_list.append(self.tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
            ref_list.append(self.tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))

        reward = [scorer(ref_list, hyp_list)[scorer_index] for scorer, scorer_index in
                  zip(self.scorers, self.scorers_index)]
        return reward, hyp_list, ref_list

    def __repr__(self):
        s = "SCST\n"
        s += json.dumps({
            'Scores': str(self.scores),
            'scores_args': str(self.scores_args),
            'scores_weights': str(self.scores_weights),
            'Generate': {'top_k': self.top_k},
        }, indent=4)
        return s



from mauve.utils import get_model, get_tokenizer, featurize_tokens_from_model
from mauve.compute_mauve import cluster_feats, get_divergence_curve_for_multinomials, compute_area_under_curve, \
    get_fronter_integral, SimpleNamespace
import numpy as np
import torch.nn as nn

def compute_mauve(
        p_features, q_features,
        num_buckets='auto', pca_max_data=-1, kmeans_explained_var=0.9,
        kmeans_num_redo=5, kmeans_max_iter=500,
        divergence_curve_discretization_size=25, mauve_scaling_factor=5,
        verbose=False, seed=25
):
    if num_buckets == 'auto':
        # heuristic: use num_clusters = num_generations / 10
        num_buckets = max(2, int(round(min(p_features.shape[0], q_features.shape[0]) / 10)))
    elif not isinstance(num_buckets, int):
        raise ValueError('num_buckets is expected to be an integer or "auto"')

    # Acutal binning
    p, q = cluster_feats(p_features, q_features,
                         num_clusters=num_buckets,
                         norm='l2', whiten=False,
                         pca_max_data=pca_max_data,
                         explained_variance=kmeans_explained_var,
                         num_redo=kmeans_num_redo,
                         max_iter=kmeans_max_iter,
                         seed=seed, verbose=verbose)
    # Divergence curve and mauve
    mixture_weights = np.linspace(1e-6, 1 - 1e-6, divergence_curve_discretization_size)
    divergence_curve = get_divergence_curve_for_multinomials(p, q, mixture_weights, mauve_scaling_factor)
    x, y = divergence_curve.T
    idxs1 = np.argsort(x)
    idxs2 = np.argsort(y)
    mauve_score = 0.5 * (
            compute_area_under_curve(x[idxs1], y[idxs1]) +
            compute_area_under_curve(y[idxs2], x[idxs2])
    )
    fi_score = get_fronter_integral(p, q)
    to_return = SimpleNamespace(
        p_hist=p, q_hist=q, divergence_curve=divergence_curve,
        mauve=mauve_score,
        frontier_integral=fi_score,
        num_buckets=num_buckets,
    )
    return to_return


class MauveScorer(nn.Module):
    def __init__(self, model, **kwargs):
        super().__init__()
        self.model = model
        self.tokenizer = get_tokenizer(self.model)
        self.model = get_model(self.model, self.tokenizer, -1)

    def forward(self, refs, hyps):
        tokenized_refs = [self.tokenizer.encode(sen, return_tensors='pt', truncation=True, max_length=512)
                          for sen in refs]
        tokenized_hyps = [self.tokenizer.encode(sen, return_tensors='pt', truncation=True, max_length=512)
                          for sen in hyps]
        p_features = featurize_tokens_from_model(self.model, tokenized_refs, name="p").detach().cpu().numpy()
        q_features = featurize_tokens_from_model(self.model, tokenized_hyps, name="q").detach().cpu().numpy()

        return compute_mauve(p_features=p_features, q_features=q_features).mauve


if __name__ == '__main__':
    m = MauveScorer(model=None or "distilgpt2")
    refs = [
        'no evidence of consolidation to suggest pneumonia is seen. there  is some retrocardiac atelectasis. a small left pleural effusion may be  present. no pneumothorax is seen. no pulmonary edema. a right granuloma is  unchanged. the heart is mildly enlarged, unchanged. there is tortuosity of  the aorta.',
        'there are moderate bilateral pleural effusions with overlying atelectasis,  underlying consolidation not excluded. mild prominence of the interstitial  markings suggests mild pulmonary edema. the cardiac silhouette is mildly  enlarged. the mediastinal contours are unremarkable. there is no evidence of  pneumothorax.'
    ]
    hyps = [
        'heart size is moderately enlarged. the mediastinal and hilar contours are unchanged. there is no pulmonary edema. small left pleural effusion is present. patchy opacities in the lung bases likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.',
        'heart size is mildly enlarged. the mediastinal and hilar contours are normal. there is mild pulmonary edema. moderate bilateral pleural effusions are present, left greater than right. bibasilar airspace opacities likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.'
    ]
    print(m.compute(refs, hyps))



import inspect
import warnings
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import torch
import torch.distributed as dist
from torch import nn

from transformers.file_utils import ModelOutput
from transformers.generation_beam_constraints import Constraint
from transformers.generation_beam_search import BeamScorer, BeamSearchScorer
from transformers.generation_logits_process import (
    EncoderNoRepeatNGramLogitsProcessor,
    ForcedBOSTokenLogitsProcessor,
    ForcedEOSTokenLogitsProcessor,
    HammingDiversityLogitsProcessor,
    InfNanRemoveLogitsProcessor,
    LogitsProcessorList,
    MinLengthLogitsProcessor,
    NoBadWordsLogitsProcessor,
    NoRepeatNGramLogitsProcessor,
    PrefixConstrainedLogitsProcessor,
    RepetitionPenaltyLogitsProcessor,
    TemperatureLogitsWarper,
    TopKLogitsWarper,
    TopPLogitsWarper,

)
from transformers.generation_stopping_criteria import (
    MaxLengthCriteria,
    MaxTimeCriteria,
    StoppingCriteria,
    StoppingCriteriaList,
    validate_stopping_criteria,
)
from transformers.pytorch_utils import torch_int_div
from transformers.utils import logging

from transformers.generation_utils import BeamSearchOutput
from transformers.generation_utils import BeamSearchEncoderDecoderOutput
from transformers.generation_utils import BeamSearchDecoderOnlyOutput


def _validate_model_kwargs(self, model_kwargs):
    return


def prepare_inputs_for_generation(self, input_ids, past=None, attention_mask=None, **kwargs):
    input_shape = input_ids.shape
    # if model is used as a decoder in encoder-decoder model, the decoder attention mask is created on the fly
    if attention_mask is None:
        attention_mask = input_ids.new_ones(input_shape)

    # cut decoder_input_ids if past is used
    if past is not None:
        input_ids = input_ids[:, -1:]
    return {"input_ids": input_ids, "attention_mask": attention_mask, "past_key_values": past, **kwargs}


def beam_search(
        self,
        input_ids: torch.LongTensor,
        beam_scorer: BeamScorer,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        max_length: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        output_scores: Optional[bool] = None,
        return_dict_in_generate: Optional[bool] = None,
        synced_gpus: Optional[bool] = False,
        **model_kwargs,
) -> Union[BeamSearchOutput, torch.LongTensor]:
    r"""
    Generates sequences for models with a language modeling head using beam search decoding.

    Parameters:

        input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
            The sequence used as a prompt for the generation.
        beam_scorer (`BeamScorer`):
            An derived instance of [`BeamScorer`] that defines how beam hypotheses are constructed, stored and
            sorted during generation. For more information, the documentation of [`BeamScorer`] should be read.
        logits_processor (`LogitsProcessorList`, *optional*):
            An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
            used to modify the prediction scores of the language modeling head applied at each generation step.
        stopping_criteria (`StoppingCriteriaList`, *optional*):
            An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
            used to tell if the generation loop should stop.
        max_length (`int`, *optional*, defaults to 20):
            **DEPRECATED**. Use `logits_processor` or `stopping_criteria` directly to cap the number of generated
            tokens. The maximum length of the sequence to be generated.
        pad_token_id (`int`, *optional*):
            The id of the *padding* token.
        eos_token_id (`int`, *optional*):
            The id of the *end-of-sequence* token.
        output_attentions (`bool`, *optional*, defaults to `False`):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under
            returned tensors for more details.
        output_hidden_states (`bool`, *optional*, defaults to `False`):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
            for more details.
        output_scores (`bool`, *optional*, defaults to `False`):
            Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
        return_dict_in_generate (`bool`, *optional*, defaults to `False`):
            Whether or not to return a [`~file_utils.ModelOutput`] instead of a plain tuple.
        synced_gpus (`bool`, *optional*, defaults to `False`):
            Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
        model_kwargs:
            Additional model specific kwargs will be forwarded to the `forward` function of the model. If model is
            an encoder-decoder model the kwargs should include `encoder_outputs`.

    Return:
        [`generation_utilsBeamSearchDecoderOnlyOutput`], [`~generation_utils.BeamSearchEncoderDecoderOutput`] or
        `torch.LongTensor`: A `torch.LongTensor` containing the generated tokens (default behaviour) or a
        [`~generation_utils.BeamSearchDecoderOnlyOutput`] if `model.config.is_encoder_decoder=False` and
        `return_dict_in_generate=True` or a [`~generation_utils.BeamSearchEncoderDecoderOutput`] if
        `model.config.is_encoder_decoder=True`.


    Examples:

    ```python
    >>> from transformers import (
    ...     AutoTokenizer,
    ...     AutoModelForSeq2SeqLM,
    ...     LogitsProcessorList,
    ...     MinLengthLogitsProcessor,
    ...     BeamSearchScorer,
    ... )
    >>> import torch

    >>> tokenizer = AutoTokenizer.from_pretrained("t5-base")
    >>> model = AutoModelForSeq2SeqLM.from_pretrained("t5-base")

    >>> encoder_input_str = "translate English to German: How old are you?"
    >>> encoder_input_ids = tokenizer(encoder_input_str, return_tensors="pt").input_ids


    >>> # lets run beam search using 3 beams
    >>> num_beams = 3
    >>> # define decoder start token ids
    >>> input_ids = torch.ones((num_beams, 1), device=model.device, dtype=torch.long)
    >>> input_ids = input_ids * model.config.decoder_start_token_id

    >>> # add encoder_outputs to model keyword arguments
    >>> model_kwargs = {
    ...     "encoder_outputs": model.get_encoder()(
    ...         encoder_input_ids.repeat_interleave(num_beams, dim=0), return_dict=True
    ...     )
    ... }

    >>> # instantiate beam scorer
    >>> beam_scorer = BeamSearchScorer(
    ...     batch_size=1,
    ...     num_beams=num_beams,
    ...     device=model.device,
    ... )

    >>> # instantiate logits processors
    >>> logits_processor = LogitsProcessorList(
    ...     [
    ...         MinLengthLogitsProcessor(5, eos_token_id=model.config.eos_token_id),
    ...     ]
    ... )

    >>> outputs = model.beam_search(input_ids, beam_scorer, logits_processor=logits_processor, **model_kwargs)

    >>> print("Generated:", tokenizer.batch_decode(outputs, skip_special_tokens=True))
    ```"""
    # init values
    logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
    stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
    if max_length is not None:
        warnings.warn(
            "`max_length` is deprecated in this function, use `stopping_criteria=StoppingCriteriaList(MaxLengthCriteria(max_length=max_length))` instead.",
            UserWarning,
        )
        stopping_criteria = validate_stopping_criteria(stopping_criteria, max_length)
    if len(stopping_criteria) == 0:
        warnings.warn("You don't have defined any stopping_criteria, this will likely loop forever", UserWarning)
    pad_token_id = pad_token_id if pad_token_id is not None else self.config.pad_token_id
    eos_token_id = eos_token_id if eos_token_id is not None else self.config.eos_token_id
    output_scores = output_scores if output_scores is not None else self.config.output_scores
    output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate is not None else self.config.return_dict_in_generate
    )

    batch_size = len(beam_scorer._beam_hyps)
    num_beams = beam_scorer.num_beams

    batch_beam_size, cur_len = input_ids.shape

    if num_beams * batch_size != batch_beam_size:
        raise ValueError(
            f"Batch dimension of `input_ids` should be {num_beams * batch_size}, but is {batch_beam_size}."
        )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    beam_indices = (
        tuple(() for _ in range(batch_beam_size)) if (return_dict_in_generate and output_scores) else None
    )
    decoder_attentions = () if (return_dict_in_generate and output_attentions) else None
    cross_attentions = () if (return_dict_in_generate and output_attentions) else None
    decoder_hidden_states = () if (return_dict_in_generate and output_hidden_states) else None

    # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
    if return_dict_in_generate and self.config.is_encoder_decoder:
        encoder_attentions = model_kwargs["encoder_outputs"].get("attentions") if output_attentions else None
        encoder_hidden_states = (
            model_kwargs["encoder_outputs"].get("hidden_states") if output_hidden_states else None
        )

    beam_scores = torch.zeros((batch_size, num_beams), dtype=torch.float, device=input_ids.device)
    beam_scores[:, 1:] = -1e9
    beam_scores = beam_scores.view((batch_size * num_beams,))
    # Added by JB
    hf_models = model_kwargs.pop("hf_models")
    model_kwargs_list = model_kwargs.pop("encoders_outputs")
    ###
    this_peer_finished = False  # used by synced_gpus only
    while True:

        if synced_gpus:
            # Under synced_gpus the `forward` call must continue until all gpus complete their sequence.
            # The following logic allows an early break if all peers finished generating their sequence
            this_peer_finished_flag = torch.tensor(0.0 if this_peer_finished else 1.0).to(input_ids.device)
            # send 0.0 if we finished, 1.0 otherwise
            dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
            # did all peers finish? the reduced sum will be 0.0 then
            if this_peer_finished_flag.item() == 0.0:
                break

        # Changed by JB
        models_inputs = [hf_models[0].prepare_inputs_for_generation(input_ids, **m) for m in model_kwargs_list]
        outputs = [hf(**model_inputs,
                      return_dict=True,
                      output_attentions=output_attentions,
                      output_hidden_states=output_hidden_states)
                   for hf, model_inputs in zip(hf_models, models_inputs)]

        if synced_gpus and this_peer_finished:
            cur_len = cur_len + 1
            continue  # don't waste resources running the code we don't need

        next_token_logits = sum(o.logits[:, -1, :] for o in outputs)
        ######

        # hack: adjust tokens for Marian. For Marian we have to make sure that the `pad_token_id`
        # cannot be generated both before and after the `nn.functional.log_softmax` operation.
        next_token_logits = self.adjust_logits_during_generation(next_token_logits, cur_len=cur_len)
        next_token_scores = nn.functional.log_softmax(
            next_token_logits, dim=-1
        )  # (batch_size * num_beams, vocab_size)

        next_token_scores_processed = logits_processor(input_ids, next_token_scores)
        next_token_scores = next_token_scores_processed + beam_scores[:, None].expand_as(next_token_scores)

        # Store scores, attentions and hidden_states when required
        if return_dict_in_generate:
            if output_scores:
                scores += (next_token_scores_processed,)
            if output_attentions:
                decoder_attentions += (
                    (outputs.decoder_attentions,) if self.config.is_encoder_decoder else (outputs.attentions,)
                )
                if self.config.is_encoder_decoder:
                    cross_attentions += (outputs.cross_attentions,)

            if output_hidden_states:
                decoder_hidden_states += (
                    (outputs.decoder_hidden_states,)
                    if self.config.is_encoder_decoder
                    else (outputs.hidden_states,)
                )

        # reshape for beam search
        vocab_size = next_token_scores.shape[-1]
        next_token_scores = next_token_scores.view(batch_size, num_beams * vocab_size)

        next_token_scores, next_tokens = torch.topk(
            next_token_scores, 2 * num_beams, dim=1, largest=True, sorted=True
        )

        next_indices = torch_int_div(next_tokens, vocab_size)
        next_tokens = next_tokens % vocab_size

        # stateless
        beam_outputs = beam_scorer.process(
            input_ids,
            next_token_scores,
            next_tokens,
            next_indices,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
        )

        beam_scores = beam_outputs["next_beam_scores"]
        beam_next_tokens = beam_outputs["next_beam_tokens"]
        beam_idx = beam_outputs["next_beam_indices"]

        input_ids = torch.cat([input_ids[beam_idx, :], beam_next_tokens.unsqueeze(-1)], dim=-1)

        # Changed by JB
        model_kwargs_list = [
            hf._update_model_kwargs_for_generation(o, m, is_encoder_decoder=True)
            for hf, o, m in zip(hf_models, outputs, model_kwargs_list)]

        for m in model_kwargs_list:
            if m["past"] is not None:
                m["past"] = self._reorder_cache(m["past"], beam_idx)
        ###

        if return_dict_in_generate and output_scores:
            beam_indices = tuple((beam_indices[beam_idx[i]] + (beam_idx[i],) for i in range(len(beam_indices))))

        # increase cur_len
        cur_len = cur_len + 1

        if beam_scorer.is_done or stopping_criteria(input_ids, scores):
            if not synced_gpus:
                break
            else:
                this_peer_finished = True

    sequence_outputs = beam_scorer.finalize(
        input_ids,
        beam_scores,
        next_tokens,
        next_indices,
        pad_token_id=pad_token_id,
        eos_token_id=eos_token_id,
        max_length=stopping_criteria.max_length,
    )

    if return_dict_in_generate:
        if not output_scores:
            sequence_outputs["sequence_scores"] = None
        else:
            num_return_sequences = beam_scorer.num_beam_hyps_to_keep
            # return only as many indices as sequences
            beam_indices = tuple(
                (beam_indices[i * num_beams: i * num_beams + num_return_sequences] for i in range(batch_size))
            )
            beam_indices = sum(beam_indices, ())

        if self.config.is_encoder_decoder:
            return BeamSearchEncoderDecoderOutput(
                sequences=sequence_outputs["sequences"],
                sequences_scores=sequence_outputs["sequence_scores"],
                scores=scores,
                beam_indices=beam_indices,
                encoder_attentions=encoder_attentions,
                encoder_hidden_states=encoder_hidden_states,
                decoder_attentions=decoder_attentions,
                cross_attentions=cross_attentions,
                decoder_hidden_states=decoder_hidden_states,
            )
        else:
            return BeamSearchDecoderOnlyOutput(
                sequences=sequence_outputs["sequences"],
                sequences_scores=sequence_outputs["sequence_scores"],
                scores=scores,
                beam_indices=beam_indices,
                attentions=decoder_attentions,
                hidden_states=decoder_hidden_states,
            )
    else:
        return sequence_outputs["sequences"]



import logging

logging.getLogger('faiss.loader').setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.ERROR)
logging.getLogger("stanza").setLevel(logging.WARNING)
logging.getLogger("filelock").setLevel(logging.CRITICAL)
logging.getLogger("numba").setLevel(logging.WARNING)
logging.getLogger("matplotlib").setLevel(logging.WARNING)
logging.getLogger('PIL').setLevel(logging.WARNING)

from .zoo.modeling_auto import AutoModel
from .models import *



import torch
import json
from random import sample

import faiss
import copy
from torch import Tensor
import math
from tqdm import tqdm
from omegaconf import OmegaConf
import numpy as np
import torch.nn as nn
from vilmedic.models.utils import get_n_params
from pl_bolts.models.self_supervised.simclr.simclr_module import Projection
from vilmedic.blocks.vision import *


def nt_xent_loss(out_1, out_2, temperature, eps=1e-6):
    """
    assume out_1 and out_2 are normalized
    out_1: [batch_size, dim]
    out_2: [batch_size, dim]
    """
    # gather representations in case of distributed training
    # out_1_dist: [batch_size * world_size, dim]
    # out_2_dist: [batch_size * world_size, dim]
    out_1_dist = out_1
    out_2_dist = out_2

    # out: [2 * batch_size, dim]
    # out_dist: [2 * batch_size * world_size, dim]
    out = torch.cat([out_1, out_2], dim=0)
    out_dist = torch.cat([out_1_dist, out_2_dist], dim=0)

    # cov and sim: [2 * batch_size, 2 * batch_size * world_size]
    # neg: [2 * batch_size]
    cov = torch.mm(out, out_dist.t().contiguous())
    sim = torch.exp(cov / temperature)
    neg = sim.sum(dim=-1)

    # from each row, subtract e^(1/temp) to remove similarity measure for x1.x1
    row_sub = Tensor(neg.shape).fill_(math.e ** (1 / temperature)).to(neg.device)
    neg = torch.clamp(neg - row_sub, min=eps)  # clamp for numerical stability

    # Positive similarity, pos becomes [2 * batch_size]
    pos = torch.exp(torch.sum(out_1 * out_2, dim=-1) / temperature)
    pos = torch.cat([pos, pos], dim=0)

    loss = -torch.log(pos / (neg + eps)).mean()

    return loss


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


def run_kmeans(x,
               num_cluster,
               temperature,
               ):
    """
    Args:
        x: data to be clustered
    """

    print('performing kmeans clustering')
    results = {'im2cluster': [], 'centroids': [], 'density': []}

    for seed, num_cluster in enumerate(num_cluster):
        # intialize faiss clustering parameters
        d = x.shape[1]
        k = int(num_cluster)
        clus = faiss.Clustering(d, k)
        clus.verbose = True
        clus.niter = 20
        clus.nredo = 5
        clus.seed = seed
        clus.max_points_per_centroid = 1000
        clus.min_points_per_centroid = 10

        res = faiss.StandardGpuResources()
        cfg = faiss.GpuIndexFlatConfig()
        cfg.useFloat16 = False
        cfg.device = 0
        index = faiss.GpuIndexFlatL2(res, d, cfg)

        clus.train(x, index)

        D, I = index.search(x, 1)  # for each sample, find cluster distance and assignments
        im2cluster = [int(n[0]) for n in I]

        # get cluster centroids
        centroids = faiss.vector_to_array(clus.centroids).reshape(k, d)

        # sample-to-centroid distances for each cluster
        Dcluster = [[] for c in range(k)]
        for im, i in enumerate(im2cluster):
            Dcluster[i].append(D[im][0])

        # concentration estimation (phi)
        density = np.zeros(k)
        for i, dist in enumerate(Dcluster):
            if len(dist) > 1:
                d = (np.asarray(dist) ** 0.5).mean() / np.log(len(dist) + 10)
                density[i] = d

                # if cluster only has one point, use the max to estimate its concentration
        dmax = density.max()
        for i, dist in enumerate(Dcluster):
            if len(dist) <= 1:
                density[i] = dmax

        density = density.clip(np.percentile(density, 10),
                               np.percentile(density, 90))  # clamp extreme values for stability
        density = temperature * density / density.mean()  # scale the mean to temperature

        # convert to cuda Tensors for broadcast
        centroids = torch.Tensor(centroids).cuda()
        centroids = nn.functional.normalize(centroids, p=2, dim=1)

        im2cluster = torch.LongTensor(im2cluster).cuda()
        density = torch.Tensor(density).cuda()

        results['centroids'].append(centroids)
        results['density'].append(density)
        results['im2cluster'].append(im2cluster)

    return results


def evaluation(models, config, dl, **kwargs):
    # No ensembling for this evaluation
    model = models[0]
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for i, batch in enumerate(pbar):
        out = model(**batch, eval_mode=True)
        visuals.append(out['visual'].cpu())

    # Give PCL class the current evaluation features
    visuals = torch.cat(visuals)
    PCL2.EVAL_FEATURES = copy.deepcopy(visuals.numpy())

    return {'loss': -1,
            'visual': visuals
            }


class PCL2(nn.Module):
    EVAL_FEATURES = None

    def __init__(self, config, clustering_epoch_start, dl, gradient_accumulation=False, **kwargs):
        super().__init__()

        self.clustering_epoch_start = clustering_epoch_start

        # Visual Encoder
        self.config = copy.deepcopy(config)

        cnn = {
            "proto": "CNN",
            "backbone": "resnet50",
            "output_layer": "avgpool",
            "dropout_out": 0.0,
            "permute": "batch_first",
            "freeze": False
        }

        self.visual = eval(cnn.pop('proto'))(**cnn)
        self.projection = Projection(input_dim=2048, hidden_dim=2048, output_dim=128)
        # Evaluation
        self.eval_func = evaluation

        #
        self.current_epoch = -1
        self.cluster_result = None

        # Loss
        self.criterion = nn.CrossEntropyLoss()
        self.gradient_accumulation = gradient_accumulation

    def forward(self, images, index, eval_mode=False, epoch=None, **kwargs):
        if eval_mode or not self.gradient_accumulation:
            return self.forward_accumulation(images, index, eval_mode=eval_mode, epoch=epoch, **kwargs)
        else:
            bs = images.shape[0]
            images = torch.chunk(images, 2, dim=0)
            index = [index[:bs // 2], index[bs // 2:]]
            loss = 0
            for image, ind in zip(images, index):
                loss += self.forward_accumulation(image, ind, eval_mode=eval_mode, epoch=epoch, **kwargs)["loss"]
            return {"loss": loss}

    def forward_accumulation(self, images, index, eval_mode=False, epoch=None, **kwargs):
        if eval_mode:
            return {"visual": self.projection(self.visual(images.cuda()))}

        # New epoch, do clustering
        if epoch + 1 >= self.clustering_epoch_start and epoch != self.current_epoch:
            assert PCL2.EVAL_FEATURES is not None
            self.cluster_result = run_kmeans(PCL2.EVAL_FEATURES,
                                             self.config.num_cluster,
                                             self.config.temperature,
                                             )  # run kmeans clustering on master node
            self.current_epoch = epoch

        images = torch.split(images.cuda(), [3, 3], dim=1)
        z1 = self.projection(self.visual(images[0]))
        z2 = self.projection(self.visual(images[1]))
        loss = nt_xent_loss(z1, z2, self.config.simclr_temperature)

        proto_labels = None
        proto_logits = None
        if self.cluster_result is not None:
            proto_labels = []
            proto_logits = []
            for n, (im2cluster, prototypes, density) in enumerate(
                    zip(self.cluster_result['im2cluster'], self.cluster_result['centroids'],
                        self.cluster_result['density'])):
                # get positive prototypes
                pos_proto_id = im2cluster[index]
                pos_prototypes = prototypes[pos_proto_id]

                # sample negative prototypes
                all_proto_id = [i for i in range(im2cluster.max())]
                neg_proto_id = set(all_proto_id) - set(pos_proto_id.tolist())
                neg_proto_id = sample(neg_proto_id, self.config.pcl_r)  # sample r negative prototypes
                neg_prototypes = prototypes[neg_proto_id]

                proto_selected = torch.cat([pos_prototypes, neg_prototypes], dim=0)

                # compute prototypical logits
                logits_proto = torch.mm(z1, proto_selected.t())

                # targets for prototype assignment
                labels_proto = torch.linspace(0, z1.size(0) - 1, steps=z1.size(0)).long().cuda()

                # scaling temperatures for the selected prototypes
                temp_proto = density[torch.cat([pos_proto_id, torch.LongTensor(neg_proto_id).cuda()], dim=0)]
                logits_proto /= temp_proto

                proto_labels.append(labels_proto)
                proto_logits.append(logits_proto)

        accp, acc = None, None
        if proto_logits is not None:
            loss_proto = 0
            for proto_out, proto_target in zip(proto_logits, proto_labels):
                loss_proto += self.criterion(proto_out, proto_target)
                accp = accuracy(proto_out, proto_target)[0]

            # average loss across all sets of prototypes
            loss_proto /= len(self.config.num_cluster)
            loss += loss_proto

        return {"loss": loss, "custom_print": "accp: {}".format(acc, accp)}

    def __repr__(self):
        s = "PCL\n"
        s += str(json.dumps(OmegaConf.to_container(self.config), indent=4)) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
import torch.nn as nn
import logging
from torchvision.models import *
from torchxrayvision.models import DenseNet as XrvDenseNet, ResNet as XrvResNet
from monai.networks.nets.densenet import densenet121 as MonaiDensenet121, densenet169 as MonaiDensenet169, \
    densenet201 as MonaiDensenet201, densenet264 as MonaiDensenet264

from .vgg_hgap import *
from transformers.models.vit.modeling_vit import ViTModel, ViTConfig
from transformers.models.deit.modeling_deit import DeiTModel, DeiTConfig
from transformers.modeling_outputs import BaseModelOutputWithPooling, BaseModelOutputWithPoolingAndNoAttention, \
    BaseModelOutputWithNoAttention
from transformers import ResNetConfig, PoolFormerConfig
from transformers.models.resnet.modeling_resnet import ResNetModel as HFResNetModel
from transformers.models.poolformer.modeling_poolformer import PoolFormerModel as HFPoolFormerModel


class _3d_densenet121(MonaiDensenet121):
    pass


class _3d_densenet169(MonaiDensenet169):
    pass


class _3d_densenet201(MonaiDensenet201):
    pass


class _3d_densenet264(MonaiDensenet264):
    pass


def get_network(backbone, output_layer, pretrained, weights=None, **kwargs):
    """
    Create sub-network given a backbone and an output_layer
    """
    # Create avgpool for densenet, does not exist as such
    if 'densenet' in backbone and '3d' not in backbone and output_layer == 'avgpool':
        sub_network = get_network(backbone, 'features', pretrained, **kwargs)
        sub_network.add_module('relu', nn.ReLU(inplace=True))
        sub_network.add_module('avgpool', nn.AdaptiveAvgPool2d((1, 1)))
        sub_network.add_module('flatten', nn.Flatten(1))
        return sub_network

    # HuggingFace Vision transformer
    if "vit" in backbone.lower():
        model = ViTModel(ViTConfig(return_dict=True, **kwargs), add_pooling_layer=False)
        model.layernorm = nn.Identity()
        return model

    if "deit" in backbone.lower():
        return DeiTModel(DeiTConfig(return_dict=True, **kwargs), add_pooling_layer=False)

    # HuggingFace ResNet
    if "hfresnet" in backbone.lower():
        return HFResNetModel(ResNetConfig(return_dict=True, **kwargs))

    # HuggingFace PoolFormer
    if "hfpoolformer" in backbone.lower():
        return HFPoolFormerModel(PoolFormerConfig(return_dict=True, **kwargs))

    # Torchxrayvision
    if "xrv" in backbone.lower():
        network = eval(backbone)(weights=weights)
        if hasattr(network, 'model'):  # XrvResNet Fix
            network = network.model
    # PyTorch
    else:
        network = eval(backbone)(pretrained=pretrained, **kwargs)

    if output_layer is not None and (not output_layer == 'classifier'):
        layers = [n for n, _ in network.named_children()]
        assert output_layer in layers, '{} not in {}'.format(output_layer, layers)
        sub_network = []
        for n, c in network.named_children():
            sub_network.append(c)
            if n == output_layer:
                break
        network = nn.Sequential(*sub_network)

    return network


class CNN(nn.Module):
    def __init__(self,
                 backbone,
                 permute,
                 dropout_out=0.0,
                 freeze=False,
                 output_layer=None,
                 pretrained=True,
                 visual_embedding_dim=None,
                 **kwargs):
        super(CNN, self).__init__()

        self.backbone = backbone
        self.output_layer = output_layer
        self.permute = permute
        self.freeze = freeze
        self.pretrained = pretrained
        self.cnn = get_network(self.backbone, self.output_layer, self.pretrained, **kwargs)
        self.dropout_out = nn.Dropout(p=dropout_out)
        self.is3D = "3d" in backbone

        assert permute in ["batch_first", "spatial_first", "no_permute"]

        if freeze:
            for name, param in self.cnn.named_parameters():
                param.requires_grad = False

    def forward(self, images, **kwargs):
        out = self.cnn(images)
        if isinstance(self.cnn, ViTModel) or isinstance(self.cnn, DeiTModel):
            assert isinstance(out, BaseModelOutputWithPooling)
            out = self.dropout_out(out.last_hidden_state)
            return out

        if isinstance(self.cnn, HFResNetModel):
            assert isinstance(out, BaseModelOutputWithPoolingAndNoAttention)
            out = out.last_hidden_state

        if isinstance(self.cnn, HFPoolFormerModel):
            assert isinstance(out, BaseModelOutputWithNoAttention)
            out = out.last_hidden_state

        out = self.dropout_out(out)
        if self.permute == "no_permute":
            out = out
        elif self.permute == "batch_first":
            out = out.view(*out.size()[:2], -1).permute(0, 2, 1)
            if out.shape[1] == 1:  # avgpool case
                out = out.squeeze(1)
        elif self.permute == "spatial_first":
            out = out.view(*out.size()[:2], -1).permute(2, 0, 1)
        else:
            raise NotImplementedError()
        return out

    def train(self, mode: bool = True):
        if self.freeze:
            mode = False
        self.training = mode
        for module in self.children():
            module.train(mode)
        return self

    def __repr__(self):
        s = str(self.backbone) + \
            '(' + \
            (str(type(self.cnn).__name__) + '(' + str(self.cnn.config) + '), ' if self.backbone.lower() in ['deit',
                                                                                                            'vit'] else '') + \
            'dropout_out=' + str(self.dropout_out.p) + \
            ', freeze=' + str(self.freeze) + \
            (', output_layer=' + str(self.output_layer) if self.output_layer is not None else '') + \
            (', pretrained=' + str(self.pretrained) if self.backbone.lower() not in ['deit', 'vit'] else '') + \
            ('\n classifier= {}'.format(list(self.cnn.children())[-1]) if self.output_layer == 'classifier' else '' + ')')
        return s



import torch
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence
from torch import nn, einsum

from einops import rearrange, repeat
from einops.layers.torch import Rearrange

# helpers

def exists(val):
    return val is not None

def pair(t):
    return t if isinstance(t, tuple) else (t, t)

# adaptive token sampling functions and classes

def log(t, eps = 1e-6):
    return torch.log(t + eps)

def sample_gumbel(shape, device, dtype, eps = 1e-6):
    u = torch.empty(shape, device = device, dtype = dtype).uniform_(0, 1)
    return -log(-log(u, eps), eps)

def batched_index_select(values, indices, dim = 1):
    value_dims = values.shape[(dim + 1):]
    values_shape, indices_shape = map(lambda t: list(t.shape), (values, indices))
    indices = indices[(..., *((None,) * len(value_dims)))]
    indices = indices.expand(*((-1,) * len(indices_shape)), *value_dims)
    value_expand_len = len(indices_shape) - (dim + 1)
    values = values[(*((slice(None),) * dim), *((None,) * value_expand_len), ...)]

    value_expand_shape = [-1] * len(values.shape)
    expand_slice = slice(dim, (dim + value_expand_len))
    value_expand_shape[expand_slice] = indices.shape[expand_slice]
    values = values.expand(*value_expand_shape)

    dim += value_expand_len
    return values.gather(dim, indices)

class AdaptiveTokenSampling(nn.Module):
    def __init__(self, output_num_tokens, eps = 1e-6):
        super().__init__()
        self.eps = eps
        self.output_num_tokens = output_num_tokens

    def forward(self, attn, value, mask):
        heads, output_num_tokens, eps, device, dtype = attn.shape[1], self.output_num_tokens, self.eps, attn.device, attn.dtype

        # first get the attention values for CLS token to all other tokens

        cls_attn = attn[..., 0, 1:]

        # calculate the norms of the values, for weighting the scores, as described in the paper

        value_norms = value[..., 1:, :].norm(dim = -1)

        # weigh the attention scores by the norm of the values, sum across all heads

        cls_attn = einsum('b h n, b h n -> b n', cls_attn, value_norms)

        # normalize to 1

        normed_cls_attn = cls_attn / (cls_attn.sum(dim = -1, keepdim = True) + eps)

        # instead of using inverse transform sampling, going to invert the softmax and use gumbel-max sampling instead

        pseudo_logits = log(normed_cls_attn)

        # mask out pseudo logits for gumbel-max sampling

        mask_without_cls = mask[:, 1:]
        mask_value = -torch.finfo(attn.dtype).max / 2
        pseudo_logits = pseudo_logits.masked_fill(~mask_without_cls, mask_value)

        # expand k times, k being the adaptive sampling number

        pseudo_logits = repeat(pseudo_logits, 'b n -> b k n', k = output_num_tokens)
        pseudo_logits = pseudo_logits + sample_gumbel(pseudo_logits.shape, device = device, dtype = dtype)

        # gumble-max and add one to reserve 0 for padding / mask

        sampled_token_ids = pseudo_logits.argmax(dim = -1) + 1

        # calculate unique using torch.unique and then pad the sequence from the right

        unique_sampled_token_ids_list = [torch.unique(t, sorted = True) for t in torch.unbind(sampled_token_ids)]
        unique_sampled_token_ids = pad_sequence(unique_sampled_token_ids_list, batch_first = True)

        # calculate the new mask, based on the padding

        new_mask = unique_sampled_token_ids != 0

        # CLS token never gets masked out (gets a value of True)

        new_mask = F.pad(new_mask, (1, 0), value = True)

        # prepend a 0 token id to keep the CLS attention scores

        unique_sampled_token_ids = F.pad(unique_sampled_token_ids, (1, 0), value = 0)
        expanded_unique_sampled_token_ids = repeat(unique_sampled_token_ids, 'b n -> b h n', h = heads)

        # gather the new attention scores

        new_attn = batched_index_select(attn, expanded_unique_sampled_token_ids, dim = 2)

        # return the sampled attention scores, new mask (denoting padding), as well as the sampled token indices (for the residual)
        return new_attn, new_mask, unique_sampled_token_ids

# classes

class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn
    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)

class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout = 0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout)
        )
    def forward(self, x):
        return self.net(x)

class Attention(nn.Module):
    def __init__(self, dim, heads = 8, dim_head = 64, dropout = 0., output_num_tokens = None):
        super().__init__()
        inner_dim = dim_head *  heads
        self.heads = heads
        self.scale = dim_head ** -0.5

        self.attend = nn.Softmax(dim = -1)
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias = False)

        self.output_num_tokens = output_num_tokens
        self.ats = AdaptiveTokenSampling(output_num_tokens) if exists(output_num_tokens) else None

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x, *, mask):
        num_tokens = x.shape[1]

        qkv = self.to_qkv(x).chunk(3, dim = -1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = self.heads), qkv)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        if exists(mask):
            dots_mask = rearrange(mask, 'b i -> b 1 i 1') * rearrange(mask, 'b j -> b 1 1 j')
            mask_value = -torch.finfo(dots.dtype).max
            dots = dots.masked_fill(~dots_mask, mask_value)

        attn = self.attend(dots)

        sampled_token_ids = None

        # if adaptive token sampling is enabled
        # and number of tokens is greater than the number of output tokens
        if exists(self.output_num_tokens) and (num_tokens - 1) > self.output_num_tokens:
            attn, mask, sampled_token_ids = self.ats(attn, v, mask = mask)

        out = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')

        return self.to_out(out), mask, sampled_token_ids

class Transformer(nn.Module):
    def __init__(self, dim, depth, max_tokens_per_depth, heads, dim_head, mlp_dim, dropout = 0.):
        super().__init__()
        assert len(max_tokens_per_depth) == depth, 'max_tokens_per_depth must be a tuple of length that is equal to the depth of the transformer'
        assert sorted(max_tokens_per_depth, reverse = True) == list(max_tokens_per_depth), 'max_tokens_per_depth must be in decreasing order'
        assert min(max_tokens_per_depth) > 0, 'max_tokens_per_depth must have at least 1 token at any layer'

        self.layers = nn.ModuleList([])
        for _, output_num_tokens in zip(range(depth), max_tokens_per_depth):
            self.layers.append(nn.ModuleList([
                PreNorm(dim, Attention(dim, output_num_tokens = output_num_tokens, heads = heads, dim_head = dim_head, dropout = dropout)),
                PreNorm(dim, FeedForward(dim, mlp_dim, dropout = dropout))
            ]))

    def forward(self, x):
        b, n, device = *x.shape[:2], x.device

        # use mask to keep track of the paddings when sampling tokens
        # as the duplicates (when sampling) are just removed, as mentioned in the paper
        mask = torch.ones((b, n), device = device, dtype = torch.bool)

        token_ids = torch.arange(n, device = device)
        token_ids = repeat(token_ids, 'n -> b n', b = b)

        for attn, ff in self.layers:
            attn_out, mask, sampled_token_ids = attn(x, mask = mask)

            # when token sampling, one needs to then gather the residual tokens with the sampled token ids
            if exists(sampled_token_ids):
                x = batched_index_select(x, sampled_token_ids, dim = 1)
                token_ids = batched_index_select(token_ids, sampled_token_ids, dim = 1)

            x = x + attn_out

            x = ff(x) + x

        return x, token_ids

class ViT(nn.Module):
    def __init__(self, *, image_size, patch_size, num_classes, dim, depth, max_tokens_per_depth, heads, mlp_dim, channels = 3, dim_head = 64, dropout = 0., emb_dropout = 0.):
        super().__init__()
        image_height, image_width = pair(image_size)
        patch_height, patch_width = pair(patch_size)

        assert image_height % patch_height == 0 and image_width % patch_width == 0, 'Image dimensions must be divisible by the patch size.'

        num_patches = (image_height // patch_height) * (image_width // patch_width)
        patch_dim = channels * patch_height * patch_width

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1 = patch_height, p2 = patch_width),
            nn.Linear(patch_dim, dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))
        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))
        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(dim, depth, max_tokens_per_depth, heads, dim_head, mlp_dim, dropout)

        self.mlp_head = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, num_classes)
        )

    def forward(self, img, return_sampled_token_ids = False):
        x = self.to_patch_embedding(img)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, '() n d -> b n d', b = b)
        x = torch.cat((cls_tokens, x), dim=1)
        x += self.pos_embedding[:, :(n + 1)]
        x = self.dropout(x)

        x, token_ids = self.transformer(x)

        logits = self.mlp_head(x[:, 0])

        if return_sampled_token_ids:
            # remove CLS token and decrement by 1 to make -1 the padding
            token_ids = token_ids[:, 1:] - 1
            return logits, token_ids

        return logits



import torch
from torch.utils.data import Dataset

from ..base.ImageDataset import ImageDataset
import os
import logging


class RCNNDataset(Dataset):
    def __init__(self, label, image, bbox, split, ckpt_dir, **kwargs):
        self.split = split
        self.image = ImageDataset(**image, split=split)

        self.bbox = open(os.path.join(bbox.root, split + '.' + bbox.file)).readlines()
        self.label = open(os.path.join(label.root, split + '.' + label.file)).readlines()

        # str to list
        self.bbox = [eval(bbox) for bbox in self.bbox]
        self.label = [eval(label) for label in self.label]
        assert len(self.image) == len(self.label) == len(self.bbox)

    def __getitem__(self, index):
        bbox = self.bbox[index]
        bbox = torch.as_tensor(bbox, dtype=torch.float32)

        labels = self.label[index]
        area = (bbox[:, 3] - bbox[:, 1]) * (bbox[:, 2] - bbox[:, 0])
        iscrowd = torch.zeros((len(bbox),), dtype=torch.int64)

        assert len(bbox) == len(labels)

        return {'image': self.image.__getitem__(index),
                'boxes': torch.as_tensor(bbox, dtype=torch.float32),
                'image_id': torch.tensor([index]),
                'area': area,
                'iscrowd': torch.zeros((len(bbox),), dtype=torch.int64),
                'labels': torch.as_tensor(iscrowd, dtype=torch.int64),
                }

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {'images': torch.stack([s['image'] for s in batch]),
                        'targets': [{"labels": s["labels"],
                                     "boxes": s["boxes"],
                                     "image_id": s["image_id"],
                                     "iscrowd": s["iscrowd"],
                                     "area": s["area"],
                                     } for s in batch]
                        }
            return collated

        return collate_fn

    def __len__(self):
        return len(self.image)

    def __repr__(self):
        return ""



import pathlib
import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 8):
    sys.exit('Sorry, Python >=3.8 is required for ViLMedic.')

setup(
    name='vilmedic',
    version='1.3.1',
    description='ViLMedic is a modular framework for multimodal research at the intersection of vision and language in the medical field.',
    author='Jean-Benoit Delbrouck',
    license='MIT',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='medical nlp deep-learning pytorch',
    python_requires='>=3.8',
    setup_requires="Cython",
    install_requires=['appdirs',
                      'omegaconf==2.0.6',
                      'urllib3==1.26.14',
                      'protobuf==3.19.6',
                      'torchvision==0.9.1',
                      'rouge_score',
                      'youtokentome==1.0.3',
                      'tokenizers==0.11.6',
                      'scikit_image==0.18.2',
                      'scikit-learn==0.24.2',
                      'pydicom==2.2.0',
                      'transformers==4.23.1',
                      'seaborn==0.11.1',
                      'dalle-pytorch==1.4.2',
                      'torchxrayvision==0.0.32',
                      'stanza==1.3.0',
                      'bert-score',
                      'torch==1.8.1',
                      'pytorch-lightning==1.4.2',
                      'pytorch-metric-learning==0.9.99',
                      'torch-optimizer==0.1.0',
                      'umap-learn',
                      'opencv-python==4.5.4.60',
                      'mauve-text',
                      'numba',
                      'torchmetrics==0.5.0',
                      'numpy==1.21',
                      'gdown==4.6.0',
                      'spacy===3.2.3',
                      'sentencepiece',
                      'radgraph==0.0.5',
                      'f1chexbert==0.0.1',
                      'psutil==5.9.0',
                      'lightning-bolts==0.5.0',
                      'faiss-gpu',
                      'pyedflib',
                      'nibabel',
                      'monai==1.1.0',
                      'typing_extensions==4.4.0'
                      ],
    include_package_data=True,
    exclude_package_data={'': ['.git']},
    packages=find_packages(exclude=["bin"]),
    scripts=[str(p) for p in pathlib.Path('bin/scripts').glob('*')],
    zip_safe=False)



# -*- coding: utf-8 -*-
import torch
import torch.nn as nn
import torch.nn.functional as F


class Attention(nn.Module):
    """Attention layer for seq2seq NMT."""
    def __init__(self, ctx_dim, hid_dim, att_bottleneck='ctx',
                 transform_ctx=True, att_activ='tanh', att_type='mlp',
                 mlp_bias=False, temp=1., ctx2hid=True):
        super().__init__()

        self.activ = getattr(torch, att_activ)
        self.ctx_dim = ctx_dim
        self.hid_dim = hid_dim
        self._ctx2hid = ctx2hid
        self.temperature = temp
        self.att_type = att_type

        # The common dimensionality for inner formulation
        if isinstance(att_bottleneck, int):
            self.mid_dim = att_bottleneck
        else:
            self.mid_dim = getattr(self, '{}_dim'.format(att_bottleneck))

        if self.att_type == 'mlp':
            self.mlp = nn.Linear(self.mid_dim, 1, bias=False)
            self._scores = self._mlp_scores
        elif self.att_type == 'dot':
            self._scores = self._dot_scores
        else:
            raise Exception('Unknown attention type {}'.format(att_type))

        # Adaptor from RNN's hidden dim to mid_dim
        self.hid2ctx = nn.Linear(self.hid_dim, self.mid_dim, bias=False)

        if transform_ctx or self.mid_dim != self.ctx_dim:
            # Additional context projection within same dimensionality
            self.ctx2ctx = nn.Linear(self.ctx_dim, self.mid_dim, bias=False)
        else:
            self.ctx2ctx = lambda x: x

        if self._ctx2hid:
            # ctx2hid: final transformation from ctx to hid
            self.ctx2hid = nn.Linear(self.ctx_dim, self.hid_dim, bias=False)
        else:
            self.ctx2hid = lambda x: x

        if mlp_bias:
            self.bias = nn.Parameter(torch.Tensor(self.mid_dim))
            self.bias.data.zero_()
        else:
            self.register_parameter('bias', None)

    def forward(self, hid, ctx, ctx_mask=None):
        r"""Computes attention probabilities and final context using
        decoder's hidden state and source annotations.

        Arguments:
            hid(Variable): A set of decoder hidden states of shape `T*B*H`
                where `T` == 1, `B` is batch dim and `H` is hidden state dim.
            ctx(Variable): A set of annotations of shape `S*B*C` where `S`
                is the source timestep dim, `B` is batch dim and `C`
                is annotation dim.
            ctx_mask(FloatTensor): A binary mask of shape `S*B` with zeroes
                in the padded positions.

        Returns:
            scores(Variable): A variable of shape `S*B` containing normalized
                attention scores for each position and sample.
            z_t(Variable): A variable of shape `B*H` containing the final
                attended context vector for this target decoding timestep.

        Notes:
            This will only work when `T==1` for now.
        """
        # scores -> SxB
        scores = self._scores(self.ctx2ctx(ctx),    # SxBxC
                              self.hid2ctx(hid))    # TxBxC

        # Normalize attention scores correctly -> S*B
        if ctx_mask is not None:
            # Mask out padded positions with -inf so that they get 0 attention
            scores.masked_fill_((1 - ctx_mask).bool(), -1e8)

        alpha = F.softmax(scores, dim=0)

        # Transform final context vector to H for further decoders
        return alpha, self.ctx2hid((alpha.unsqueeze(-1) * ctx).sum(0))

    def _dot_scores(self, ctx_, hid_):
        # shuffle dims to prepare for batch mat-mult
        return torch.bmm(hid_.permute(1, 0, 2), ctx_.permute(1, 2, 0)).div(
            self.temperature).squeeze(1).t()

    def _mlp_scores(self, ctx_, hid_):
        inner_sum = ctx_ + hid_
        if self.bias is not None:
            inner_sum += self.bias
        return self.mlp(self.activ(inner_sum)).div(self.temperature).squeeze(-1)



import torch.nn as nn
from rouge_score import rouge_scorer
from six.moves import zip_longest
import numpy as np


class Rouge(nn.Module):
    def __init__(self, rouges, **kwargs):
        super().__init__()
        rouges = [r.replace('rougel', 'rougeL') for r in rouges]
        self.scorer = rouge_scorer.RougeScorer(rouges, use_stemmer=True)
        self.rouges = rouges

    def forward(self, refs, hyps):
        scores = []
        for target_rec, prediction_rec in zip_longest(refs, hyps):
            if target_rec is None or prediction_rec is None:
                raise ValueError("Must have equal number of lines across target and "
                                 "prediction.")
            scores.append(self.scorer.score(target_rec, prediction_rec))

        # aggregator = scoring.BootstrapAggregator()
        # for score in scores:
        #     aggregator.add_scores(score)
        # print(aggregator.aggregate())
        f1_rouge = [s[self.rouges[0]].fmeasure for s in scores]
        return np.mean(f1_rouge), f1_rouge


class Rouge1(Rouge):
    def __init__(self, **kwargs):
        super(Rouge1, self).__init__(rouges=['rouge1'])


class Rouge2(Rouge):
    def __init__(self, **kwargs):
        super(Rouge2, self).__init__(rouges=['rouge2'])


class RougeL(Rouge):
    def __init__(self, **kwargs):
        super(RougeL, self).__init__(rouges=['rougeL'])



import torch
import torch.nn as nn
from vilmedic.blocks.vision import *
from vilmedic.blocks.classifier import *
from vilmedic.blocks.classifier.evaluation import evaluation
from vilmedic.blocks.losses import *

from vilmedic.models.utils import get_n_params

from transformers.models.bert.modeling_bert import BertEncoder, BertPooler
from transformers.models.bert_generation import BertGenerationConfig


class MVQA(nn.Module):
    def __init__(self, cnn, classifier, adapter, transformer, loss, **kwargs):
        super(MVQA, self).__init__()

        cnn_func = cnn.pop('proto')
        loss_func = loss.pop('proto')
        classifier_func = classifier.pop('proto')

        self.cnn = eval(cnn_func)(**cnn)
        self.adapter = nn.Sequential(
            nn.Linear(adapter.pop('input_size'), adapter.pop('output_size')),
            torch.nn.LayerNorm(transformer.hidden_size, eps=transformer.layer_norm_eps)
        )

        bert_conf = BertGenerationConfig(**transformer)
        self.transformer = BertEncoder(bert_conf)
        self.pooler = BertPooler(bert_conf)

        self.classifier = eval(classifier_func)(**classifier)

        self.loss_func = eval(loss_func)(**loss).cuda()

        print(self.loss_func)
        # Evaluation
        self.eval_func = evaluation

    def forward(self, images, labels=None, from_training=True, iteration=None, epoch=None, **kwargs):
        out = self.cnn(images.cuda())
        out = self.adapter(out)
        out = self.transformer(out, output_attentions=True)

        attentions = out.attentions  # num_layers, batch_size, num_heads, sequence_length, sequence_length

        out = self.pooler(out.last_hidden_state)
        out = self.classifier(out)

        loss = torch.tensor(0.)
        if from_training:
            loss = self.loss_func(out, labels.cuda(), **kwargs)

        return {'loss': loss, 'output': out, 'answer': torch.argmax(out, dim=-1)}

    def __repr__(self):
        s = super().__repr__() + '\n'
        s += "{}\n".format(get_n_params(self))
        return s
